"""
Test job-script files.
"""

from pathlib import Path
from textwrap import dedent

import pytest

from jobbergate_api.apps.applications.application_files import ApplicationFiles
from jobbergate_api.apps.job_scripts.job_script_files import (
    JOBSCRIPTS_MAIN_FILE_FOLDER,
    JOBSCRIPTS_SUPPORTING_FILES_FOLDER,
    JOBSCRIPTS_WORK_DIR,
    JobScriptFiles,
    inject_sbatch_params,
)


@pytest.fixture
def job_script_data_as_string():
    """
    Provide a fixture that returns an example of a default application script.
    """
    content = dedent(
        """
        #!/bin/bash

        #SBATCH --job-name=rats
        #SBATCH --partition=debug
        #SBATCH --output=sample-%j.out


        echo $SLURM_TASKS_PER_NODE
        echo $SLURM_SUBMIT_DIR
        """
    ).strip()
    return content


@pytest.fixture
def new_job_script_data_as_string():
    """
    Provide a fixture that returns an application script after the injection of the sbatch params.
    """
    content = dedent(
        """
        #!/bin/bash

        #SBATCH --job-name=rats
        #SBATCH --comment=some_comment
        #SBATCH --nice=-1
        #SBATCH -N 10
        #SBATCH --partition=debug
        #SBATCH --output=sample-%j.out


        echo $SLURM_TASKS_PER_NODE
        echo $SLURM_SUBMIT_DIR
        """
    ).strip()
    return content


@pytest.fixture
def sbatch_params():
    """
    Provide a fixture that returns string content of the argument --sbatch-params.
    """
    return ["--comment=some_comment", "--nice=-1", "-N 10"]


def test_inject_sbatch_params(job_script_data_as_string, sbatch_params, new_job_script_data_as_string):
    """
    Test the injection of sbatch params in a default application script.
    """
    injected_string = inject_sbatch_params(job_script_data_as_string, sbatch_params)
    assert injected_string == new_job_script_data_as_string


@pytest.fixture
def dummy_application_files(
    dummy_application_source_file,
    dummy_application_config,
    dummy_template,
):
    """
    Test fixture with dummy application files.
    """
    return ApplicationFiles(
        templates={"test_job_script.sh": dummy_template},
        source_file=dummy_application_source_file,
        config_file=dummy_application_config,
    )


class TestJobScriptFiles:
    """
    Test JobScriptFiles.
    """

    def test_check_main_file_path_is_in_files_keys__success(self):
        """
        Test a case were JobScriptFiles is created with success.
        """
        jobscript_files = JobScriptFiles(
            main_file_path="jobbergate.sh", files={"jobbergate.sh": "dummy-file-content"}
        )

        assert isinstance(jobscript_files, JobScriptFiles)

    def test_check_main_file_path_is_in_files_keys__error_main_file_path(self):
        """
        Test that ValueError is raised when the main file is not included in the files.
        """
        with pytest.raises(ValueError):
            JobScriptFiles(
                main_file_path="wrong-filename",
                files={"jobbergate.sh": "dummy-file-content"},
            )

    def test_get_from_s3(self, mocked_file_manager_factory):
        """
        Test that JobScriptFiles can be obtained from the file manager.
        """
        job_script_id = 1

        work_dir = mocked_file_manager_factory / JOBSCRIPTS_WORK_DIR / str(job_script_id)
        output_dir = Path("test-output-dir")

        main_file_name = output_dir / "jobbergate.sh"
        main_file_content = "test-main-file"
        main_file_path = work_dir / JOBSCRIPTS_MAIN_FILE_FOLDER / main_file_name
        main_file_path.parent.mkdir(parents=True, exist_ok=True)
        main_file_path.write_text(main_file_content)

        supporting_file_name = output_dir / "support.sh"
        supporting_file_content = "test-support-file"
        supporting_file_path = work_dir / JOBSCRIPTS_SUPPORTING_FILES_FOLDER / supporting_file_name
        supporting_file_path.parent.mkdir(parents=True, exist_ok=True)
        supporting_file_path.write_text(supporting_file_content)

        desired_jobscript_files = JobScriptFiles(
            main_file_path=main_file_name,
            files={
                main_file_name: main_file_content,
                supporting_file_name: supporting_file_content,
            },
        )

        actual_jobscript_files = JobScriptFiles.get_from_s3(job_script_id)

        assert desired_jobscript_files == actual_jobscript_files

    def test_get_from_s3__no_main_file(self, mocked_file_manager_factory):
        """
        Test that ValueError is raised when no main file is found.
        """
        job_script_id = 1
        with pytest.raises(
            ValueError,
            match="One and only one main file is expected for a job-script, found 0",
        ):
            JobScriptFiles.get_from_s3(job_script_id)

    def test_delete_from_s3(self, mocked_file_manager_factory):
        """
        Test that JobScriptFiles are deleted as expected.
        """
        job_script_id = 1

        work_dir = mocked_file_manager_factory / JOBSCRIPTS_WORK_DIR / str(job_script_id)
        output_dir = Path("test-output-dir")

        main_file_name = output_dir / "jobbergate.sh"
        main_file_content = "test-main-file"
        main_file_path = work_dir / JOBSCRIPTS_MAIN_FILE_FOLDER / main_file_name
        main_file_path.parent.mkdir(parents=True, exist_ok=True)
        main_file_path.write_text(main_file_content)

        supporting_file_name = output_dir / "support.sh"
        supporting_file_content = "test-support-file"
        supporting_file_path = work_dir / JOBSCRIPTS_SUPPORTING_FILES_FOLDER / supporting_file_name
        supporting_file_path.parent.mkdir(parents=True, exist_ok=True)
        supporting_file_path.write_text(supporting_file_content)

        assert main_file_path.is_file()
        assert supporting_file_path.is_file()

        JobScriptFiles.delete_from_s3(job_script_id)

        assert not main_file_path.is_file()
        assert not supporting_file_path.is_file()

    def test_write_to_s3(self, mocked_file_manager_factory):
        """
        Test that JobScriptFiles are written as expected.
        """
        job_script_id = 1

        work_dir = mocked_file_manager_factory / JOBSCRIPTS_WORK_DIR / str(job_script_id)
        output_dir = Path("test-output-dir")

        main_file_name = output_dir / "jobbergate.sh"
        main_file_content = "test-main-file"
        main_file_path = work_dir / JOBSCRIPTS_MAIN_FILE_FOLDER / main_file_name

        supporting_file_name = output_dir / "support.sh"
        supporting_file_content = "test-support-file"
        supporting_file_path = work_dir / JOBSCRIPTS_SUPPORTING_FILES_FOLDER / supporting_file_name

        JobScriptFiles(
            main_file_path=main_file_name,
            files={
                main_file_name: main_file_content,
                supporting_file_name: supporting_file_content,
            },
        ).write_to_s3(job_script_id)

        assert main_file_path.is_file()
        assert main_file_path.read_text() == main_file_content

        assert supporting_file_path.is_file()
        assert supporting_file_path.read_text() == supporting_file_content

    def test_io_integration_with_s3(self, mocked_file_manager_factory):
        """
        End-to-end test where the JobScriptFiles are written and read from s3.
        """
        job_script_id = 1
        output_dir = Path("test-output-dir")

        main_file_name = output_dir / "jobbergate.sh"
        main_file_content = "test-main-file"

        supporting_file_name = output_dir / "support.sh"
        supporting_file_content = "test-support-file"

        actual_jobscript_files = JobScriptFiles(
            main_file_path=main_file_name,
            files={
                main_file_name: main_file_content,
                supporting_file_name: supporting_file_content,
            },
        )

        actual_jobscript_files.write_to_s3(job_script_id)

        desired_jobscript_files = JobScriptFiles.get_from_s3(job_script_id)

        assert desired_jobscript_files == actual_jobscript_files

        JobScriptFiles.delete_from_s3(job_script_id)

        with pytest.raises(
            ValueError,
            match="One and only one main file is expected for a job-script, found 0",
        ):
            JobScriptFiles.get_from_s3(job_script_id)

    def test_render_from_application__success(
        self,
        param_dict,
        job_script_data_as_string,
        dummy_application_files,
    ):
        """
        Test that JobScriptFiles can be obtained when rendering ApplicationFiles.
        """
        desired_job_script_files = JobScriptFiles(
            main_file_path="test_job_script.sh",
            files={
                "test_job_script.sh": job_script_data_as_string,
                "support_file_b.py": job_script_data_as_string,
            },
        )

        actual_job_script_files = JobScriptFiles.render_from_application(
            dummy_application_files,
            param_dict,
        )

        assert actual_job_script_files == desired_job_script_files

    def test_render_from_application__ignores_template_path_prefix_in_default_template_file(
        self,
        param_dict,
        job_script_data_as_string,
        dummy_application_files,
    ):
        """
        Test that JobScriptFiles can be obtained when rendering ApplicationFiles.
        """
        mangled_param_dict = param_dict.copy()
        mangled_param_dict["jobbergate_config"]["default_template"] = "templates/test_job_script.sh"

        desired_job_script_files = JobScriptFiles(
            main_file_path="test_job_script.sh",
            files={
                "test_job_script.sh": job_script_data_as_string,
                "support_file_b.py": job_script_data_as_string,
            },
        )

        actual_job_script_files = JobScriptFiles.render_from_application(
            dummy_application_files,
            mangled_param_dict,
        )

        assert actual_job_script_files == desired_job_script_files



"""
Router for the JobScript resource.
"""
from typing import Optional

from armasec import TokenPayload
from fastapi import APIRouter, Depends, HTTPException, Query, status
from loguru import logger

from jobbergate_api.apps.applications.application_files import ApplicationFiles
from jobbergate_api.apps.applications.models import applications_table
from jobbergate_api.apps.applications.schemas import ApplicationResponse
from jobbergate_api.apps.job_scripts.job_script_files import JobScriptCreationError, JobScriptFiles
from jobbergate_api.apps.job_scripts.models import job_scripts_table, searchable_fields, sortable_fields
from jobbergate_api.apps.job_scripts.schemas import (
    JobScriptCreateRequest,
    JobScriptPartialResponse,
    JobScriptResponse,
    JobScriptUpdateRequest,
)
from jobbergate_api.apps.permissions import Permissions
from jobbergate_api.pagination import Pagination, ok_response, package_response
from jobbergate_api.security import IdentityClaims, guard
from jobbergate_api.storage import (
    INTEGRITY_CHECK_EXCEPTIONS,
    database,
    render_sql,
    search_clause,
    sort_clause,
)

router = APIRouter()


@router.post(
    "/job-scripts",
    status_code=status.HTTP_201_CREATED,
    response_model=JobScriptResponse,
    description="Endpoint for job_script creation",
)
async def job_script_create(
    job_script: JobScriptCreateRequest,
    token_payload: TokenPayload = Depends(guard.lockdown(Permissions.JOB_SCRIPTS_EDIT)),
):
    """
    Create a new job script.

    Make a post request to this endpoint with the required values to create a new job script.
    """
    logger.debug(f"Creating {job_script=}")

    select_query = applications_table.select().where(applications_table.c.id == job_script.application_id)
    logger.trace(f"select_query = {render_sql(select_query)}")

    raw_application = await database.fetch_one(select_query)

    if not raw_application:
        message = f"Application with id={job_script.application_id} not found."
        logger.warning(message)
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail=message)

    application = ApplicationResponse.parse_obj(raw_application)

    if application.application_uploaded is False:
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail=f"Application with id={job_script.application_id} was not uploaded.",
        )

    identity_claims = IdentityClaims.from_token_payload(token_payload)

    create_dict = dict(
        **{k: v for (k, v) in job_script.dict(exclude_unset=True).items() if k != "param_dict"},
        job_script_owner_email=identity_claims.email,
    )

    try:
        jobscript_files = JobScriptFiles.render_from_application(
            application_files=ApplicationFiles.get_from_s3(application.id),
            user_supplied_parameters=job_script.param_dict,
            sbatch_params=create_dict.pop("sbatch_params", []),
        )
    except JobScriptCreationError as e:
        message = str(e)
        logger.error(message)
        raise HTTPException(status_code=status.HTTP_422_UNPROCESSABLE_ENTITY, detail=message)

    logger.debug("Inserting job_script")

    async with database.transaction():

        try:
            insert_query = job_scripts_table.insert().returning(job_scripts_table)
            logger.trace(f"insert_query = {render_sql(insert_query)}")
            job_script_data = await database.fetch_one(query=insert_query, values=create_dict)

            if job_script_data is None:
                message = "An error occurred when inserting the JobScript at the database."
                logger.error(message)
                raise HTTPException(
                    status_code=status.HTTP_400_BAD_REQUEST,
                    detail=message,
                )

            jobscript_files.write_to_s3(job_script_data["id"])

        except INTEGRITY_CHECK_EXCEPTIONS as e:
            logger.error(f"Reverting database transaction: {str(e)}")
            raise HTTPException(status_code=status.HTTP_422_UNPROCESSABLE_ENTITY, detail=str(e))

        except KeyError as e:
            logger.error(f"Reverting database transaction: {str(e)}")
            raise HTTPException(status_code=status.HTTP_422_UNPROCESSABLE_ENTITY, detail=str(e))

    logger.debug(f"Job-script created: {dict(job_script_data)}")

    response = JobScriptResponse(
        **job_script_data,
        job_script_files=jobscript_files,
    )
    return response


@router.get(
    "/job-scripts/{job_script_id}",
    description="Endpoint to get a job_script",
    response_model=JobScriptResponse,
    dependencies=[Depends(guard.lockdown(Permissions.JOB_SCRIPTS_VIEW))],
)
async def job_script_get(job_script_id: int = Query(...)):
    """
    Return the job_script given its id.
    """
    logger.debug(f"Getting {job_script_id=}")

    query = job_scripts_table.select().where(job_scripts_table.c.id == job_script_id)
    logger.trace(f"get_query = {render_sql(query)}")
    job_script = await database.fetch_one(query)

    if not job_script:
        message = f"JobScript with id={job_script_id} not found."
        logger.warning(message)
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=message,
        )

    try:
        response = JobScriptResponse(
            **job_script,
            job_script_files=JobScriptFiles.get_from_s3(job_script_id),
        )
    except (KeyError, ValueError):
        message = f"JobScript file not found for id={job_script_id}."
        logger.warning(message)
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=message,
        )

    logger.debug(f"Job-script data: {response}")

    return response


@router.get(
    "/job-scripts",
    description="Endpoint to list job_scripts",
    responses=ok_response(JobScriptPartialResponse),
)
async def job_script_list(
    pagination: Pagination = Depends(),
    all: Optional[bool] = Query(False),
    search: Optional[str] = Query(None),
    sort_field: Optional[str] = Query(None),
    sort_ascending: bool = Query(True),
    token_payload: TokenPayload = Depends(guard.lockdown(Permissions.JOB_SCRIPTS_VIEW)),
):
    """
    List job_scripts for the authenticated user.

    Note::

       Use responses instead of response_model to skip a second round of validation and serialization. This
       is already happening in the ``package_response`` method. So, we uses ``responses`` so that FastAPI
       can generate the correct OpenAPI spec but not post-process the response.
    """
    logger.debug("Preparing to list job-scripts")

    query = job_scripts_table.select()
    identity_claims = IdentityClaims.from_token_payload(token_payload)
    if not all:
        query = query.where(job_scripts_table.c.job_script_owner_email == identity_claims.email)
    if search is not None:
        query = query.where(search_clause(search, searchable_fields))
    if sort_field is not None:
        query = query.order_by(sort_clause(sort_field, sortable_fields, sort_ascending))

    logger.trace(f"Query = {render_sql(query)}")
    return await package_response(JobScriptPartialResponse, query, pagination)


@router.delete(
    "/job-scripts/{job_script_id}",
    status_code=status.HTTP_204_NO_CONTENT,
    description="Endpoint to delete job script",
    dependencies=[Depends(guard.lockdown(Permissions.JOB_SCRIPTS_EDIT))],
)
async def job_script_delete(job_script_id: int = Query(..., description="id of the job script to delete")):
    """
    Delete job_script given its id.
    """
    logger.debug(f"Preparing to delete {job_script_id=}")
    where_stmt = job_scripts_table.c.id == job_script_id

    get_query = job_scripts_table.select().where(where_stmt)
    logger.trace(f"get_query = {render_sql(get_query)}")

    raw_job_script = await database.fetch_one(get_query)
    if not raw_job_script:

        message = f"JobScript with id={job_script_id} not found."
        logger.warning(message)
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=message,
        )

    delete_query = job_scripts_table.delete().where(where_stmt)
    logger.trace(f"delete_query = {render_sql(delete_query)}")
    await database.execute(delete_query)

    try:
        JobScriptFiles.delete_from_s3(job_script_id)
    except KeyError:
        # There is no need to raise an error if we try to delete a file that does not exist
        logger.warning(f"Tried to delete {job_script_id=}, but it was not found on S3.")


@router.put(
    "/job-scripts/{job_script_id}",
    status_code=status.HTTP_200_OK,
    description="Endpoint to update a job_script given the id",
    response_model=JobScriptResponse,
    dependencies=[Depends(guard.lockdown(Permissions.JOB_SCRIPTS_EDIT))],
)
async def job_script_update(job_script_id: int, job_script: JobScriptUpdateRequest):
    """
    Update a job_script given its id.
    """
    logger.debug(f"Updating {job_script_id=}")

    update_query = (
        job_scripts_table.update()
        .where(job_scripts_table.c.id == job_script_id)
        .values(
            job_script.dict(exclude_unset=True, exclude={"job_script_files"}),
        )
        .returning(job_scripts_table)
    )
    logger.trace(f"update_query = {render_sql(update_query)}")

    async with database.transaction():

        try:
            result = await database.fetch_one(update_query)
        except INTEGRITY_CHECK_EXCEPTIONS as e:
            logger.error(f"Reverting database transaction: {str(e)}")
            raise HTTPException(status_code=status.HTTP_409_CONFLICT, detail=str(e))

        if result is None:
            message = f"JobScript with id={job_script_id} not found."
            logger.warning(message)
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND,
                detail=message,
            )

        job_script_response = dict(result)

        try:
            if job_script.job_script_files:
                job_script.job_script_files.write_to_s3(job_script_id)
                job_script_response["job_script_files"] = job_script.job_script_files
            else:
                job_script_response["job_script_files"] = JobScriptFiles.get_from_s3(job_script_id)
        except (KeyError, ValueError) as e:
            logger.error(f"Reverting database transaction: {str(e)}")
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND,
                detail=f"File for JobScript with id={job_script_id} not found in S3.",
            )

    return job_script_response


def include_router(app):
    """
    Include the router for this module in the app.
    """
    app.include_router(router)



"""
Provide a convenience class for managing application files.
"""

from pathlib import Path, PurePath
from typing import Dict, List, Optional, cast

from fastapi import UploadFile
from file_storehouse import FileManager
from loguru import logger
from pydantic import BaseModel, Field

from jobbergate_api.apps.applications.constants import (
    APPLICATION_CONFIG_FILE_NAME,
    APPLICATION_SOURCE_FILE_NAME,
    APPLICATION_TEMPLATE_FOLDER,
    APPLICATIONS_WORK_DIR,
)
from jobbergate_api.apps.applications.file_validation import perform_all_checks_on_uploaded_files
from jobbergate_api.config import settings
from jobbergate_api.s3_manager import IO_TRANSFORMATIONS, file_manager_factory, s3_client


def _get_template_files(file_manager: FileManager) -> Dict[str, str]:
    """
    Filter the template files out of the file manager.
    """
    template_paths = filter(lambda path: str(path.parent) == APPLICATION_TEMPLATE_FOLDER, file_manager.keys())

    return {path.name: file_manager[path] for path in template_paths}


class ApplicationFiles(BaseModel):
    """
    Model containing application files.
    """

    config_file: Optional[str] = Field(None, alias="application_config")
    source_file: Optional[str] = Field(None, alias="application_source_file")
    templates: Dict[str, str] = Field(default_factory=dict, alias="application_templates")

    class Config:
        allow_population_by_field_name = True

    @classmethod
    def get_from_s3(cls, application_id: int):
        """
        Alternative method to initialize the model getting the objects from S3.
        """
        logger.debug(f"Getting application files from S3: {application_id=}")
        file_manager = cls.file_manager_factory(application_id)

        logger.debug(
            f"The number of files available for {application_id=} is {len(file_manager)}. "
            f"They are: {', '.join(map(str, file_manager.keys()))}"
        )

        application_files = cls(
            config_file=file_manager.get(APPLICATION_CONFIG_FILE_NAME),
            source_file=file_manager.get(APPLICATION_SOURCE_FILE_NAME),
            templates=_get_template_files(file_manager),
        )

        if not application_files.config_file:
            logger.warning(f"Application config file was not found for {application_id=}")
        if not application_files.source_file:
            logger.warning(f"Application source file was not found for {application_id=}")
        if not application_files.templates:
            logger.warning(f"No template file was found for {application_id=}")

        logger.debug("Success getting application files from S3")

        return application_files

    @classmethod
    def delete_from_s3(cls, application_id: int):
        """
        Delete the files associated with the given id.
        """
        file_manager = cls.file_manager_factory(application_id)
        logger.debug(
            f"Deleting from S3 the files associated to {application_id=}. "
            f"Files to be deleted: {', '.join(map(str, file_manager.keys()))}"
        )
        file_manager.clear()
        logger.debug(f"Files were deleted for {application_id=}")

    def write_to_s3(self, application_id: int, *, remove_previous_files: bool = True):
        """
        Write to s3 the files associated with a given id.
        """
        logger.debug(f"Writing the application files to S3: {application_id=}")

        file_manager = self.file_manager_factory(application_id)

        if remove_previous_files and file_manager:
            self.delete_from_s3(application_id)

        if self.config_file:
            path = Path(APPLICATION_CONFIG_FILE_NAME)
            file_manager[path] = self.config_file

        if self.source_file:
            path = Path(APPLICATION_SOURCE_FILE_NAME)
            file_manager[path] = self.source_file

        for name, content in self.templates.items():
            path = Path(APPLICATION_TEMPLATE_FOLDER, name)
            file_manager[path] = content

        logger.debug(f"Files were written for {application_id=}")

    @classmethod
    def get_from_upload_files(cls, upload_files: List[UploadFile]):
        """
        Initialize the model getting the objects from a list of uploaded files.
        """
        logger.debug("Getting application files from the uploaded files")

        perform_all_checks_on_uploaded_files(upload_files)

        application_files = cls()

        for upload in upload_files:
            file_data = upload.file.read().decode("utf-8")
            upload.file.seek(0)
            if upload.filename == APPLICATION_SOURCE_FILE_NAME:
                application_files.source_file = file_data
            elif upload.filename == APPLICATION_CONFIG_FILE_NAME:
                application_files.config_file = file_data
            elif upload.filename.endswith((".j2", ".jinja2")):
                filename = PurePath(upload.filename).name
                application_files.templates[filename] = file_data

        logger.debug("Success getting application files from the uploaded files")

        return application_files

    @classmethod
    def file_manager_factory(self, application_id: int) -> FileManager:
        """
        Build an application file manager.
        """
        return cast(
            FileManager,
            file_manager_factory(
                id=application_id,
                s3_client=s3_client,
                bucket_name=settings.S3_BUCKET_NAME,
                work_directory=Path(APPLICATIONS_WORK_DIR),
                manager_cls=FileManager,
                transformations=IO_TRANSFORMATIONS,
            ),
        )



"""
Pytest helpers to use in all apps.
"""
from pytest import fixture

from jobbergate_api.apps.job_submissions.constants import JobSubmissionStatus


@fixture
def application_data():
    """
    Provide a fixture that supplies test application data.
    """
    return {
        "application_owner_email": "test@email.com",
        "application_name": "test_name",
    }


@fixture
def fill_application_data(application_data):
    """
    Combine user supplied application data with defaults. If there are overlaps, use the user supplied data.
    """

    def _helper(**fields):
        return {
            **application_data,
            **fields,
        }

    return _helper


@fixture
def fill_all_application_data(fill_application_data):
    """
    Combine many fields of user supplied application data with defaults.
    """

    def _helper(*all_fields):
        return [fill_application_data(**f) for f in all_fields]

    return _helper


@fixture
def job_script_data():
    """
    Provide a fixture that supplies test job_script data.
    """
    return {
        "job_script_name": "test_name",
        "job_script_owner_email": "owner1@org.com",
    }


@fixture
def fill_job_script_data(job_script_data):
    """
    Combine user supplied job_script data with defaults. If there are overlaps, use the user supplied data.
    """

    def _helper(**fields):
        return {
            **job_script_data,
            **fields,
        }

    return _helper


@fixture
def fill_all_job_script_data(fill_job_script_data):
    """
    Combine many fields of user supplied job_script data with defaults.
    """

    def _helper(*all_fields):
        return [fill_job_script_data(**f) for f in all_fields]

    return _helper


@fixture
def job_submission_data():
    """
    Provide a fixture that supplies test job_submission data.
    """
    return {
        "job_submission_name": "test_name",
        "job_submission_owner_email": "owner1@org.com",
        "client_id": "dummy-client-id",
        "status": JobSubmissionStatus.CREATED,
    }


@fixture
def fill_job_submission_data(job_submission_data):
    """
    Combine user supplied job_script data with defaults. If there are overlaps, use the user supplied data.
    """

    def _helper(**fields):
        return {
            **job_submission_data,
            **fields,
        }

    return _helper


@fixture
def fill_all_job_submission_data(fill_job_submission_data):
    """
    Combine many fields of user supplied job_submission data with defaults.
    """

    def _helper(*all_fields):
        return [fill_job_submission_data(**f) for f in all_fields]

    return _helper



"""
Test the components used to validate the uploaded files.
"""

import pytest
from fastapi import UploadFile

from jobbergate_api.apps.applications.file_validation import (
    get_suffix,
    is_valid_jinja2_template,
    is_valid_python_file,
    is_valid_yaml_file,
)


@pytest.mark.parametrize(
    "filename, suffix",
    [
        ("jobbergate.py", ".py"),
        ("jobbergate.yaml", ".yaml"),
        ("jobbergate.py.jinja2", ".jinja2"),
        ("jobbergate.py.j2", ".j2"),
    ],
)
def test_get_suffix(filename, suffix, make_dummy_file):
    """
    Test if the file suffix is correctly from an UploadFile.
    """
    dummy_file = make_dummy_file(filename)
    dummy_upload = UploadFile(filename, file=dummy_file)
    assert get_suffix(dummy_upload) == suffix


@pytest.mark.parametrize(
    "is_valid, source_code",
    [
        (False, "for i in range(10):\nprint(i)"),
        (True, "for i in range(10):\n    print(i)"),
        # Notice it does not catch counter as an unknown variable
        (True, "for i in range(counter):\n    print(i)"),
    ],
)
def test_is_valid_python_file(is_valid, source_code):
    """
    Test if a given python source code is correctly checked as valid or not.
    """
    assert is_valid_python_file(source_code) is is_valid


@pytest.mark.parametrize(
    "is_valid, yaml_file",
    [
        (False, "unbalanced blackets: ]["),
        (True, "balanced blackets: []"),
    ],
)
def test_is_valid_yaml_file(is_valid, yaml_file):
    """
    Test if a given YAML file is correctly checked as valid or not.
    """
    assert is_valid_yaml_file(yaml_file) is is_valid


@pytest.mark.parametrize(
    "is_valid, template",
    [
        (False, "Hello {{ name }!"),
        (True, "Hello {{ name }}!"),
    ],
)
def test_is_valid_jinja2_template(is_valid, template):
    """
    Test if a given python source code is correctly checked as valid or not.
    """
    assert is_valid_jinja2_template(template) is is_valid



"""
Tests for the /job-scripts/ endpoint.
"""
import json
import pathlib
from textwrap import dedent
from unittest import mock

import asyncpg
import pytest
from fastapi import status

from jobbergate_api.apps.applications.application_files import ApplicationFiles
from jobbergate_api.apps.applications.models import applications_table
from jobbergate_api.apps.job_scripts.job_script_files import JobScriptFiles
from jobbergate_api.apps.job_scripts.models import job_scripts_table
from jobbergate_api.apps.job_scripts.schemas import JobScriptPartialResponse, JobScriptResponse
from jobbergate_api.apps.permissions import Permissions
from jobbergate_api.storage import database


@pytest.fixture
def job_script_data_as_string():
    """
    Provide a fixture that returns an example of a default application script.
    """
    content = dedent(
        """
                #!/bin/bash

                #SBATCH --job-name=rats
                #SBATCH --partition=debug
                #SBATCH --output=sample-%j.out


                echo $SLURM_TASKS_PER_NODE
                echo $SLURM_SUBMIT_DIR
                """
    ).strip()
    return content


@pytest.fixture
def new_job_script_data_as_string():
    """
    Provide a fixture that returns an application script after the injection of the sbatch params.
    """
    content = dedent(
        """
                #!/bin/bash

                #SBATCH --comment=some_comment
                #SBATCH --nice=-1
                #SBATCH -N 10
                #SBATCH --job-name=rats
                #SBATCH --partition=debug
                #SBATCH --output=sample-%j.out


                echo $SLURM_TASKS_PER_NODE
                echo $SLURM_SUBMIT_DIR
                """
    ).strip()
    return content


@pytest.fixture
def sbatch_params():
    """
    Provide a fixture that returns string content of the argument --sbatch-params.
    """
    return ["--comment=some_comment", "--nice=-1", "-N 10"]


@pytest.mark.asyncio
@database.transaction(force_rollback=True)
@mock.patch("jobbergate_api.apps.job_scripts.job_script_files.JobScriptFiles.write_to_s3")
async def test_create_job_script(
    mocked_write_job_script_files_to_s3,
    fill_application_data,
    job_script_data,
    fill_job_script_data,
    param_dict,
    client,
    inject_security_header,
    time_frame,
    dummy_application_config,
    dummy_application_source_file,
    dummy_template,
    job_script_data_as_string,
    mocked_file_manager_factory,
):
    """
    Test POST /job_scripts/ correctly creates a job_script.

    This test proves that a job_script is successfully created via a POST request to the /job-scripts/
    endpoint. We show this by asserting that the job_script is created in the database after the post
    request is made, the correct status code (201) is returned.
    """
    inserted_application_id = await database.execute(
        query=applications_table.insert(),
        values=fill_application_data(
            application_owner_email="owner1@org.com",
            application_uploaded=True,
        ),
    )

    ApplicationFiles(
        templates={"test_job_script.sh": dummy_template},
        source_file=dummy_application_source_file,
        config_file=dummy_application_config,
    ).write_to_s3(inserted_application_id)

    inject_security_header("owner1@org.com", Permissions.JOB_SCRIPTS_EDIT)
    with time_frame() as window:
        response = await client.post(
            "/jobbergate/job-scripts/",
            json=fill_job_script_data(
                application_id=inserted_application_id,
                param_dict=param_dict,
            ),
        )

    assert response.status_code == status.HTTP_201_CREATED

    id_rows = await database.fetch_all("SELECT id FROM job_scripts")
    assert len(id_rows) == 1

    job_script = JobScriptResponse(**response.json())

    mocked_write_job_script_files_to_s3.assert_called_once_with(job_script.id)

    assert job_script.id == id_rows[0][0]
    assert job_script.job_script_name == job_script_data["job_script_name"]
    assert job_script.job_script_owner_email == "owner1@org.com"
    assert job_script.job_script_description is None
    assert job_script.job_script_files.main_file
    assert job_script.application_id == inserted_application_id
    assert job_script.created_at in window
    assert job_script.updated_at in window
    assert job_script.job_script_files.main_file == job_script_data_as_string


@pytest.mark.asyncio
@database.transaction(force_rollback=True)
async def test_create_job_script_application_not_uploaded(
    fill_application_data,
    fill_job_script_data,
    param_dict,
    client,
    inject_security_header,
):
    """
    Test that it is not possible to create job_script when the application is marked as not uploaded.
    """
    inserted_application_id = await database.execute(
        query=applications_table.insert(),
        values=fill_application_data(
            application_owner_email="owner1@org.com",
            application_uploaded=False,
        ),
    )

    inject_security_header("owner1@org.com", Permissions.JOB_SCRIPTS_EDIT)
    response = await client.post(
        "/jobbergate/job-scripts/",
        json=fill_job_script_data(
            application_id=inserted_application_id,
            param_dict=param_dict,
        ),
    )

    assert response.status_code == status.HTTP_422_UNPROCESSABLE_ENTITY

    count = await database.fetch_all("SELECT COUNT(*) FROM job_scripts")
    assert count[0][0] == 0


@pytest.mark.asyncio
@database.transaction(force_rollback=True)
async def test_create_job_script_bad_permission(
    fill_application_data,
    fill_job_script_data,
    param_dict,
    client,
    inject_security_header,
):
    """
    Test that it is not possible to create job_script without proper permission.

    This test proves that is not possible to create a job_script without the proper permission.
    We show this by trying to create a job_script without a permission that allow "create" then assert
    that the job_script still does not exists in the database, and the correct status code (403) is returned.
    and that the boto3 method is never called.
    """
    inserted_application_id = await database.execute(
        query=applications_table.insert(),
        values=fill_application_data(application_owner_email="owner1@org.com"),
    )

    inject_security_header("owner1@org.com", "INVALID_PERMISSION")
    response = await client.post(
        "/jobbergate/job-scripts/",
        json=fill_job_script_data(
            application_id=inserted_application_id,
            param_dict=param_dict,
        ),
    )

    assert response.status_code == status.HTTP_403_FORBIDDEN

    count = await database.fetch_all("SELECT COUNT(*) FROM job_scripts")
    assert count[0][0] == 0


@pytest.mark.asyncio
@database.transaction(force_rollback=True)
async def test_create_job_script_without_application(
    fill_job_script_data,
    param_dict,
    client,
    inject_security_header,
):
    """
    Test that is not possible to create a job_script without an application.

    This test proves that is not possible to create a job_script without an existing application.
    We show this by trying to create a job_script without an application created before, then assert that the
    job_script still does not exists in the database and the correct status code (404) is returned.
    """
    inject_security_header("owner1@org.com", Permissions.JOB_SCRIPTS_EDIT)

    response = await client.post(
        "/jobbergate/job-scripts/",
        json=fill_job_script_data(application_id=9999, param_dict=param_dict),
    )

    assert response.status_code == status.HTTP_404_NOT_FOUND

    count = await database.fetch_all("SELECT COUNT(*) FROM job_scripts")
    assert count[0][0] == 0


@database.transaction(force_rollback=True)
@pytest.mark.asyncio
@mock.patch("jobbergate_api.apps.applications.application_files.ApplicationFiles.get_from_s3")
async def test_create_job_script_file_not_found(
    mocked_get_application_files_from_s3,
    fill_application_data,
    fill_job_script_data,
    param_dict,
    client,
    inject_security_header,
):
    """
    Test that is not possible to create a job_script if the application is in the database but not in S3.

    This test proves that is not possible to create a job_script with an existing application in the
    database but not in S3, this covers for when for some reason the application file in S3 is deleted but it
    remains in the database. We show this by trying to create a job_script with an existing application that
    is not in S3 (raises BotoCoreError), then assert that the job_script still does not exists in the
    database, the correct status code (404) is returned and that the boto3 method was called.
    """
    inserted_application_id = await database.execute(
        query=applications_table.insert(),
        values=fill_application_data(
            application_owner_email="owner1@org.com",
            application_uploaded=True,
        ),
    )

    mocked_get_application_files_from_s3.return_value = ApplicationFiles()

    inject_security_header("owner1@org.com", Permissions.JOB_SCRIPTS_EDIT)
    response = await client.post(
        "/jobbergate/job-scripts/",
        json=fill_job_script_data(
            application_id=inserted_application_id,
            param_dict=param_dict,
        ),
    )

    assert response.status_code == status.HTTP_422_UNPROCESSABLE_ENTITY

    count = await database.fetch_all("SELECT COUNT(*) FROM job_scripts")
    assert count[0][0] == 0


@database.transaction(force_rollback=True)
@pytest.mark.asyncio
async def test_create_job_script_unable_to_write_file_to_s3(
    fill_application_data,
    fill_job_script_data,
    param_dict,
    client,
    inject_security_header,
    dummy_application_config,
    dummy_application_source_file,
    dummy_template,
    mocked_file_manager_factory,
):
    """
    Test that a job script is not added to the database when S3 manager gets an error.
    """
    inserted_application_id = await database.execute(
        query=applications_table.insert(),
        values=fill_application_data(
            application_owner_email="owner1@org.com",
            application_uploaded=True,
        ),
    )

    ApplicationFiles(
        templates={"test_job_script.sh": dummy_template},
        source_file=dummy_application_source_file,
        config_file=dummy_application_config,
    ).write_to_s3(inserted_application_id)

    actual_id_rows = await database.fetch_all("SELECT id FROM job_scripts")

    inject_security_header("owner1@org.com", Permissions.JOB_SCRIPTS_EDIT)
    with mock.patch(
        "jobbergate_api.apps.job_scripts.job_script_files.JobScriptFiles.write_to_s3",
        side_effect=KeyError,
    ):
        response = await client.post(
            "/jobbergate/job-scripts/",
            json=fill_job_script_data(
                application_id=inserted_application_id,
                param_dict=param_dict,
            ),
        )

    assert response.status_code == status.HTTP_422_UNPROCESSABLE_ENTITY

    desired_id_rows = await database.fetch_all("SELECT id FROM job_scripts")

    assert actual_id_rows == desired_id_rows


@pytest.mark.asyncio
@database.transaction(force_rollback=True)
async def test_get_job_script_by_id(
    client,
    fill_application_data,
    job_script_data,
    fill_job_script_data,
    inject_security_header,
    mocked_file_manager_factory,
):
    """
    Test GET /job-scripts/<id>.

    This test proves that GET /job-scripts/<id> returns the correct job-script, owned by
    the user making the request. We show this by asserting that the job_script data
    returned in the response is equal to the job_script data that exists in the database
    for the given job_script id.
    """
    inserted_application_id = await database.execute(
        query=applications_table.insert(),
        values=fill_application_data(application_owner_email="owner1@org.com"),
    )
    inserted_job_script_id = await database.execute(
        query=job_scripts_table.insert(),
        values=fill_job_script_data(application_id=inserted_application_id),
    )

    main_file_path = pathlib.Path("jobbergate.py")
    dummy_job_script_files = JobScriptFiles(
        main_file_path=main_file_path, files={main_file_path: "print(__name__)"}
    )
    dummy_job_script_files.write_to_s3(inserted_job_script_id)

    count = await database.fetch_all("SELECT COUNT(*) FROM job_scripts")
    assert count[0][0] == 1

    inject_security_header("owner1@org.com", Permissions.JOB_SCRIPTS_VIEW)

    response = await client.get(f"/jobbergate/job-scripts/{inserted_job_script_id}")

    assert response.status_code == status.HTTP_200_OK

    data = response.json()
    assert data["id"] == inserted_job_script_id
    assert data["job_script_name"] == job_script_data["job_script_name"]
    assert data["job_script_owner_email"] == "owner1@org.com"
    assert data["application_id"] == inserted_application_id
    assert JobScriptFiles(**data["job_script_files"]) == dummy_job_script_files


@pytest.mark.asyncio
@database.transaction(force_rollback=True)
async def test_get_job_script_by_id_file_not_found_at_s3(
    client,
    fill_application_data,
    fill_job_script_data,
    inject_security_header,
    mocked_file_manager_factory,
):
    """
    Test if 404 is returned if a jobscript exists in the database but the file was not found in S3.
    """
    inserted_application_id = await database.execute(
        query=applications_table.insert(),
        values=fill_application_data(application_owner_email="owner1@org.com"),
    )
    inserted_job_script_id = await database.execute(
        query=job_scripts_table.insert(),
        values=fill_job_script_data(application_id=inserted_application_id),
    )

    inject_security_header("owner1@org.com", Permissions.JOB_SCRIPTS_VIEW)

    response = await client.get(f"/jobbergate/job-scripts/{inserted_job_script_id}")

    assert response.status_code == status.HTTP_404_NOT_FOUND
    assert "JobScript file not found for id=" in response.text


@pytest.mark.asyncio
@database.transaction(force_rollback=True)
async def test_get_job_script_by_id_invalid(client, inject_security_header):
    """
    Test the correct response code is returned when a job_script does not exist.

    This test proves that GET /job-script/<id> returns the correct response code when the
    requested job_script does not exist. We show this by asserting that the status code
    returned is what we would expect given the job_script requested doesn't exist (404).
    """
    inject_security_header("owner1@org.com", Permissions.JOB_SCRIPTS_VIEW)
    response = await client.get("/jobbergate/job-scripts/9999")
    assert response.status_code == status.HTTP_404_NOT_FOUND


@pytest.mark.asyncio
@database.transaction(force_rollback=True)
async def test_get_job_script_by_id_bad_permission(
    client,
    fill_application_data,
    fill_job_script_data,
    inject_security_header,
):
    """
    Test the correct response code is returned when the user don't have the proper permission.

    This test proves that GET /job-script/<id> returns the correct response code when the
    user don't have the proper permission. We show this by asserting that the status code
    returned is what we would expect (403).
    """
    inserted_application_id = await database.execute(
        query=applications_table.insert(),
        values=fill_application_data(application_owner_email="owner1@org.com"),
    )
    inserted_job_script_id = await database.execute(
        query=job_scripts_table.insert(),
        values=fill_job_script_data(application_id=inserted_application_id),
    )
    inject_security_header("owner1@org.com", "INVALID_PERMISSION")
    response = await client.get(f"/jobbergate/job-scripts/{inserted_job_script_id}")
    assert response.status_code == status.HTTP_403_FORBIDDEN


@pytest.mark.asyncio
@database.transaction(force_rollback=True)
async def test_get_job_script__no_params(
    client,
    fill_application_data,
    fill_all_job_script_data,
    inject_security_header,
):
    """
    Test GET /job-scripts/ returns only job_scripts owned by the user making the request.

    This test proves that GET /job-scripts/ returns the correct job_scripts for the user making
    the request. We show this by asserting that the job_scripts returned in the response are
    only job_scripts owned by the user making the request.
    """
    inserted_application_id = await database.execute(
        query=applications_table.insert(),
        values=fill_application_data(application_owner_email="owner1@org.com"),
    )
    await database.execute_many(
        query=job_scripts_table.insert(),
        values=fill_all_job_script_data(
            dict(
                job_script_name="js1",
                job_script_owner_email="owner1@org.com",
                application_id=inserted_application_id,
            ),
            dict(
                job_script_name="js2",
                job_script_owner_email="owner999@org.com",
                application_id=inserted_application_id,
            ),
            dict(
                job_script_name="js3",
                job_script_owner_email="owner1@org.com",
                application_id=inserted_application_id,
            ),
        ),
    )

    count = await database.fetch_all("SELECT COUNT(*) FROM job_scripts")
    assert count[0][0] == 3

    inject_security_header("owner1@org.com", Permissions.JOB_SCRIPTS_VIEW)
    response = await client.get("/jobbergate/job-scripts/")
    assert response.status_code == status.HTTP_200_OK

    data = response.json()
    results = data.get("results")
    assert results
    assert [d["job_script_name"] for d in results] == ["js1", "js3"]

    pagination = data.get("pagination")
    assert pagination == dict(
        total=2,
        start=None,
        limit=None,
    )


@pytest.mark.asyncio
@database.transaction(force_rollback=True)
async def test_get_job_scripts__bad_permission(
    client,
    fill_application_data,
    fill_job_script_data,
    inject_security_header,
):
    """
    Test GET /job-scripts/ returns 403 since the user don't have the proper permission.

    This test proves that GET /job-scripts/ returns the 403 status code when the user making the request
    don't have the permission to list. We show this by asserting that the response status code is 403.
    """
    inserted_application_id = await database.execute(
        query=applications_table.insert(),
        values=fill_application_data(application_owner_email="owner1@org.com"),
    )
    await database.execute(
        query=job_scripts_table.insert(),
        values=fill_job_script_data(application_id=inserted_application_id),
    )

    count = await database.fetch_all("SELECT COUNT(*) FROM job_scripts")
    assert count[0][0] == 1

    inject_security_header("owner1@org.com", "INVALID_PERMISSION")
    response = await client.get("/jobbergate/job-scripts/")
    assert response.status_code == status.HTTP_403_FORBIDDEN


@pytest.mark.asyncio
@database.transaction(force_rollback=True)
async def test_get_job_scripts__with_all_param(
    client,
    fill_application_data,
    fill_all_job_script_data,
    inject_security_header,
):
    """
    Test that listing job_scripts, when all=True, contains job_scripts owned by other users.

    This test proves that the user making the request can see job_scripts owned by other users.
    We show this by creating three job_scripts, one that are owned by the user making the request, and two
    owned by another user. Assert that the response to GET /job-scripts/?all=True includes all three
    job_scripts.
    """
    inserted_application_id = await database.execute(
        query=applications_table.insert(),
        values=fill_application_data(application_owner_email="owner1@org.com"),
    )
    await database.execute_many(
        query=job_scripts_table.insert(),
        values=fill_all_job_script_data(
            {
                "job_script_name": "script1",
                "job_script_owner_email": "owner1@org.com",
                "application_id": inserted_application_id,
            },
            {
                "job_script_name": "script2",
                "job_script_owner_email": "owner999@org.com",
                "application_id": inserted_application_id,
            },
            {
                "job_script_name": "script3",
                "job_script_owner_email": "owner1@org.com",
                "application_id": inserted_application_id,
            },
        ),
    )

    count = await database.fetch_all("SELECT COUNT(*) FROM job_scripts")
    assert count[0][0] == 3

    inject_security_header("owner1@org.com", Permissions.JOB_SCRIPTS_VIEW)
    response = await client.get("/jobbergate/job-scripts/?all=True")
    assert response.status_code == status.HTTP_200_OK

    data = response.json()
    results = data.get("results")
    assert results
    assert [d["job_script_name"] for d in results] == ["script1", "script2", "script3"]

    pagination = data.get("pagination")
    assert pagination == dict(
        total=3,
        start=None,
        limit=None,
    )


@pytest.mark.asyncio
@database.transaction(force_rollback=True)
async def test_get_job_scripts__with_search_param(client, inject_security_header, fill_application_data):
    """
    Test that listing job scripts, when search=<search terms>, returns matches.

    This test proves that the user making the request will be shown job scripts that match the search string.
    We show this by creating job scripts and using various search queries to match against them.

    Assert that the response to GET /job_scripts?search=<search terms> includes correct matches.
    """
    inserted_application_id = await database.execute(
        query=applications_table.insert(),
        values=fill_application_data(application_owner_email="owner1@org.com"),
    )
    common = dict(application_id=inserted_application_id)
    await database.execute_many(
        query=job_scripts_table.insert(),
        values=[
            dict(
                id=1,
                job_script_name="test name one",
                job_script_owner_email="one@org.com",
                **common,
            ),
            dict(
                id=2,
                job_script_name="test name two",
                job_script_owner_email="two@org.com",
                **common,
            ),
            dict(
                id=22,
                job_script_name="test name twenty-two",
                job_script_owner_email="twenty-two@org.com",
                job_script_description="a long description of this job_script",
                **common,
            ),
        ],
    )
    count = await database.fetch_all("SELECT COUNT(*) FROM job_scripts")
    assert count[0][0] == 3

    inject_security_header("admin@org.com", Permissions.JOB_SCRIPTS_VIEW)

    response = await client.get("/jobbergate/job-scripts?all=true&search=one")
    assert response.status_code == status.HTTP_200_OK
    data = response.json()
    results = data.get("results")
    assert [d["id"] for d in results] == [1]

    response = await client.get("/jobbergate/job-scripts?all=true&search=two")
    assert response.status_code == status.HTTP_200_OK
    data = response.json()
    results = data.get("results")
    assert [d["id"] for d in results] == [2, 22]

    response = await client.get("/jobbergate/job-scripts?all=true&search=long")
    assert response.status_code == status.HTTP_200_OK
    data = response.json()
    results = data.get("results")
    assert [d["id"] for d in results] == [22]

    response = await client.get("/jobbergate/job-scripts?all=true&search=name+test")
    assert response.status_code == status.HTTP_200_OK
    data = response.json()
    results = data.get("results")
    assert [d["id"] for d in results] == [1, 2, 22]


@pytest.mark.asyncio
@database.transaction(force_rollback=True)
async def test_get_job_scripts__with_sort_params(
    client,
    fill_application_data,
    inject_security_header,
):
    """
    Test that listing job_scripts with sort params returns correctly ordered matches.

    This test proves that the user making the request will be shown job_scripts sorted in the correct order
    according to the ``sort_field`` and ``sort_ascending`` parameters.
    We show this by creating job_scripts and using various sort parameters to order them.

    Assert that the response to GET /job_scripts?sort_field=<field>&sort_ascending=<bool> includes correctly
    sorted job_script.
    """
    inserted_application_id = await database.execute(
        query=applications_table.insert(),
        values=fill_application_data(application_owner_email="admin@org.com"),
    )
    common = dict(
        job_script_owner_email="admin@org.com",
        application_id=inserted_application_id,
    )
    await database.execute_many(
        query=job_scripts_table.insert(),
        values=[
            dict(
                job_script_name="Z",
                **common,
            ),
            dict(
                job_script_name="Y",
                **common,
            ),
            dict(
                job_script_name="X",
                **common,
            ),
        ],
    )
    count = await database.fetch_all("SELECT COUNT(*) FROM job_scripts")
    assert count[0][0] == 3

    inject_security_header("admin@org.com", Permissions.JOB_SCRIPTS_VIEW)

    response = await client.get("/jobbergate/job-scripts?sort_field=id")
    assert response.status_code == status.HTTP_200_OK
    data = response.json()
    results = data.get("results")
    assert [d["job_script_name"] for d in results] == ["Z", "Y", "X"]

    response = await client.get("/jobbergate/job-scripts?sort_field=id&sort_ascending=false")
    assert response.status_code == status.HTTP_200_OK
    data = response.json()
    results = data.get("results")
    assert [d["job_script_name"] for d in results] == ["X", "Y", "Z"]

    response = await client.get("/jobbergate/job-scripts?sort_field=job_script_name")
    assert response.status_code == status.HTTP_200_OK
    data = response.json()
    results = data.get("results")
    assert [d["job_script_name"] for d in results] == ["X", "Y", "Z"]

    response = await client.get("/jobbergate/job-scripts?all=true&sort_field=job_script_data_as_string")
    assert response.status_code == status.HTTP_400_BAD_REQUEST
    assert "Invalid sorting column requested" in response.text


@pytest.mark.asyncio
@database.transaction(force_rollback=True)
async def test_get_job_scripts__with_pagination(
    client,
    fill_application_data,
    fill_job_script_data,
    inject_security_header,
):
    """
    Test that listing job_scripts works with pagination.

    This test proves that the user making the request can see job_scripts paginated.
    We show this by creating three job_scripts and assert that the response is correctly paginated.
    """
    inserted_application_id = await database.execute(
        query=applications_table.insert(),
        values=fill_application_data(application_owner_email="owner1@org.com"),
    )
    await database.execute_many(
        query=job_scripts_table.insert(),
        values=[
            fill_job_script_data(
                job_script_name=f"script{i}",
                job_script_owner_email="owner1@org.com",
                application_id=inserted_application_id,
            )
            for i in range(1, 6)
        ],
    )

    count = await database.fetch_all("SELECT COUNT(*) FROM job_scripts")
    assert count[0][0] == 5

    inject_security_header("owner1@org.com", Permissions.JOB_SCRIPTS_VIEW)
    response = await client.get("/jobbergate/job-scripts?start=0&limit=1")
    assert response.status_code == status.HTTP_200_OK

    data = response.json()
    results = data.get("results")
    assert results
    assert [d["job_script_name"] for d in results] == ["script1"]

    pagination = data.get("pagination")
    assert pagination == dict(total=5, start=0, limit=1)

    response = await client.get("/jobbergate/job-scripts?start=1&limit=2")
    assert response.status_code == status.HTTP_200_OK

    data = response.json()
    results = data.get("results")
    assert results
    assert [d["job_script_name"] for d in results] == ["script3", "script4"]

    pagination = data.get("pagination")
    assert pagination == dict(total=5, start=1, limit=2)

    response = await client.get("/jobbergate/job-scripts?start=2&limit=2")
    assert response.status_code == status.HTTP_200_OK

    data = response.json()
    results = data.get("results")
    assert results
    assert [d["job_script_name"] for d in results] == ["script5"]

    pagination = data.get("pagination")
    assert pagination == dict(total=5, start=2, limit=2)


@pytest.mark.freeze_time
@pytest.mark.asyncio
@database.transaction(force_rollback=True)
async def test_update_job_script(
    client,
    fill_application_data,
    fill_job_script_data,
    inject_security_header,
    time_frame,
    mocked_file_manager_factory,
):
    """
    Test update job_script via PUT.

    This test proves that the job_script values are correctly updated following a PUT request to the
    /job-scripts/<id> endpoint. We show this by assert the response status code to 201, the response data
    corresponds to the updated data, and the data in the database is also updated.
    """
    inserted_application_id = await database.execute(
        query=applications_table.insert(),
        values=fill_application_data(application_owner_email="owner1@org.com"),
    )
    inserted_job_script_id = await database.execute(
        query=job_scripts_table.insert(),
        values=fill_job_script_data(application_id=inserted_application_id),
    )

    main_file_path = "jobbergate.py"
    main_file_content = "print(__name__)"
    dummy_job_script_files = JobScriptFiles(
        main_file_path=main_file_path, files={main_file_path: main_file_content}
    )

    inject_security_header("owner1@org.com", Permissions.JOB_SCRIPTS_EDIT)
    with time_frame() as window:
        response = await client.put(
            f"/jobbergate/job-scripts/{inserted_job_script_id}",
            json={
                "job_script_name": "new name",
                "job_script_description": "new description",
                "job_script_files": {
                    "main_file_path": main_file_path,
                    "files": {main_file_path: main_file_content},
                },
            },
        )

    assert response.status_code == status.HTTP_200_OK
    data = response.json()

    assert data["job_script_name"] == "new name"
    assert data["job_script_description"] == "new description"
    assert JobScriptFiles(**data["job_script_files"]) == dummy_job_script_files
    assert JobScriptFiles.get_from_s3(inserted_job_script_id) == dummy_job_script_files
    assert data["id"] == inserted_job_script_id

    query = job_scripts_table.select(job_scripts_table.c.id == inserted_job_script_id)
    job_script = JobScriptPartialResponse.parse_obj(await database.fetch_one(query))

    assert job_script is not None
    assert job_script.job_script_name == "new name"
    assert job_script.job_script_description == "new description"
    assert job_script.updated_at in window


@pytest.mark.asyncio
@database.transaction(force_rollback=True)
async def test_update_job_script_not_found(
    client,
    inject_security_header,
):
    """
    Test that it is not possible to update a job_script not found.

    This test proves that it is not possible to update a job_script if it is not found. We show this by
    asserting that the response status code of the request is 404.
    """
    inject_security_header("owner1@org.com", Permissions.JOB_SCRIPTS_EDIT)
    response = await client.put("/jobbergate/job-scripts/123", json={"job_script_name": "new name"})

    assert response.status_code == status.HTTP_404_NOT_FOUND


@database.transaction(force_rollback=True)
@pytest.mark.asyncio
async def test_update_job_script_unable_to_write_file_to_s3(
    fill_application_data,
    fill_job_script_data,
    client,
    inject_security_header,
    mocked_file_manager_factory,
):
    """
    Test that a job script is not updated to the database when S3 manager gets an error.
    """
    inserted_application_id = await database.execute(
        query=applications_table.insert(),
        values=fill_application_data(application_owner_email="owner1@org.com"),
    )

    inserted_job_script_id = await database.execute(
        query=job_scripts_table.insert(),
        values=fill_job_script_data(application_id=inserted_application_id),
    )

    query = job_scripts_table.select(job_scripts_table.c.id == inserted_job_script_id)

    desired_jobscript_data = JobScriptPartialResponse.parse_obj(await database.fetch_one(query))

    inject_security_header("owner1@org.com", Permissions.JOB_SCRIPTS_EDIT)

    with mock.patch(
        "jobbergate_api.apps.job_scripts.job_script_files.JobScriptFiles.write_to_s3",
        side_effect=KeyError,
    ):
        response = await client.put(
            f"/jobbergate/job-scripts/{inserted_job_script_id}",
            json={
                "job_script_name": "new name",
                "job_script_description": "new description",
                "job_script_data_as_string": "new value",
            },
        )

    assert response.status_code == status.HTTP_404_NOT_FOUND
    assert "not found in S3" in response.text

    actual_jobscript_data = JobScriptPartialResponse.parse_obj(await database.fetch_one(query))

    assert desired_jobscript_data.json() == actual_jobscript_data.json()


@pytest.mark.asyncio
@database.transaction(force_rollback=True)
async def test_update_job_script_bad_permission(
    client,
    fill_application_data,
    fill_job_script_data,
    inject_security_header,
):
    """
    Test that it is not possible to update a job_script if the user don't have the proper permission.

    This test proves that it is not possible to update a job_script if the user don't have permission. We
    show this by asserting that the response status code of the request is 403, and that the data stored in
    the database for the job_script is not updated.
    """
    inserted_application_id = await database.execute(
        query=applications_table.insert(),
        values=fill_application_data(application_owner_email="owner1@org.com"),
    )
    await database.execute(
        query=job_scripts_table.insert(),
        values=fill_job_script_data(job_script_name="target-js", application_id=inserted_application_id),
    )

    inject_security_header("owner1@org.com", "INVALID_PERMISSION")
    response = await client.put("/jobbergate/job-scripts/1", data={"job_script_name": "new name"})

    assert response.status_code == status.HTTP_403_FORBIDDEN

    query = job_scripts_table.select(job_scripts_table.c.job_script_name == "target-js")
    job_script_row = await database.fetch_one(query)

    assert job_script_row is not None
    assert job_script_row["job_script_name"] == "target-js"


@pytest.mark.asyncio
@database.transaction(force_rollback=True)
async def test_delete_job_script(
    client,
    fill_application_data,
    fill_job_script_data,
    inject_security_header,
):
    """
    Test delete job_script via DELETE.

    This test proves that a job_script is successfully deleted via a DELETE request to the /job-scripts/<id>
    endpoint. We show this by asserting that the job_script no longer exists in the database after the
    request is made and the correct status code is returned (204).
    """
    inserted_application_id = await database.execute(
        query=applications_table.insert(),
        values=fill_application_data(),
    )
    inserted_job_script_id = await database.execute(
        query=job_scripts_table.insert(),
        values=fill_job_script_data(application_id=inserted_application_id),
    )

    count = await database.fetch_all("SELECT COUNT(*) FROM job_scripts")
    assert count[0][0] == 1

    with mock.patch(
        "jobbergate_api.apps.job_scripts.job_script_files.JobScriptFiles.delete_from_s3",
    ) as mocked_delete:
        inject_security_header("owner1@org.com", Permissions.JOB_SCRIPTS_EDIT)
        response = await client.delete(f"/jobbergate/job-scripts/{inserted_job_script_id}")

    assert response.status_code == status.HTTP_204_NO_CONTENT

    mocked_delete.assert_called_once()

    count = await database.fetch_all("SELECT COUNT(*) FROM job_scripts")
    assert count[0][0] == 0


@pytest.mark.asyncio
@database.transaction(force_rollback=True)
async def test_delete_job_script_not_found(client, inject_security_header):
    """
    Test that it is not possible to delete a job_script that is not found.

    This test proves that it is not possible to delete a job_script if it does not exists. We show this by
    assert that a 404 response status code is returned.
    """
    inject_security_header("owner1@org.com", Permissions.JOB_SCRIPTS_EDIT)
    response = await client.delete("/jobbergate/job-scripts/9999")

    assert response.status_code == status.HTTP_404_NOT_FOUND


@pytest.mark.asyncio
@database.transaction(force_rollback=True)
async def test_delete_job_script_bad_permission(
    client,
    fill_application_data,
    fill_job_script_data,
    inject_security_header,
):
    """
    Test that it is not possible to delete a job_script when the user don't have the permission.

    This test proves that it is not possible to delete a job_script if the user don't have the permission.
    We show this by assert that a 403 response status code is returned and the job_script still exists in
    the database after the request.
    """
    inserted_application_id = await database.execute(
        query=applications_table.insert(),
        values=fill_application_data(),
    )
    inserted_job_script_id = await database.execute(
        query=job_scripts_table.insert(),
        values=fill_job_script_data(application_id=inserted_application_id),
    )

    count = await database.fetch_all("SELECT COUNT(*) FROM job_scripts")
    assert count[0][0] == 1

    inject_security_header("owner1@org.com", "INVALID_PERMISSION")
    response = await client.delete(f"/jobbergate/job-scripts/{inserted_job_script_id}")

    assert response.status_code == status.HTTP_403_FORBIDDEN

    count = await database.fetch_all("SELECT COUNT(*) FROM job_scripts")
    assert count[0][0] == 1


@pytest.mark.asyncio
@database.transaction(force_rollback=True)
async def test_delete_job_script__fk_error(
    client,
    fill_application_data,
    fill_job_script_data,
    inject_security_header,
):
    """
    Test DELETE /job_script/<id> correctly returns a 409 on a foreign-key constraint error.
    """
    inserted_application_id = await database.execute(
        query=applications_table.insert(),
        values=fill_application_data(),
    )
    inserted_job_script_id = await database.execute(
        query=job_scripts_table.insert(),
        values=fill_job_script_data(application_id=inserted_application_id),
    )

    count = await database.fetch_all("SELECT COUNT(*) FROM job_scripts")
    assert count[0][0] == 1

    inject_security_header("owner1@org.com", Permissions.JOB_SCRIPTS_EDIT)
    with mock.patch(
        "jobbergate_api.storage.database.execute",
        side_effect=asyncpg.exceptions.ForeignKeyViolationError(
            f"""
            update or delete on table "job_scripts" violates foreign key constraint
            "job_submissions_job_script_id_fkey" on table "job_submissions"
            DETAIL:  Key (id)=({inserted_job_script_id}) is still referenced from table "job_submissions".
            """
        ),
    ):
        response = await client.delete(f"/jobbergate/job-scripts/{inserted_job_script_id}")
    assert response.status_code == status.HTTP_409_CONFLICT
    error_data = json.loads(response.text)["detail"]
    assert error_data["message"] == "Delete failed due to foreign-key constraint"
    assert error_data["table"] == "job_submissions"
    assert error_data["pk_id"] == f"{inserted_job_script_id}"



"""
Test the storage module.
"""

import enum

import pytest
from sqlalchemy import Column, Enum, Integer, Table

from jobbergate_api.metadata import metadata
from jobbergate_api.storage import build_db_url, database, sort_clause


def test_build_db_url__creates_database_url_from_parts(tweak_settings):
    """
    Provide a test ase for the ``build_db_url()`` function.

    Tests that the build_db_url function computes a database url value from separate
    DATABASE_ settings when the DEPLOY_ENV is not TEST.
    """
    with tweak_settings(
        DEPLOY_ENV="LOCAL",
        DATABASE_USER="built-base-user",
        DATABASE_PSWD="built-base-pswd",
        DATABASE_HOST="built-base-host",
        DATABASE_NAME="built-base-name",
        DATABASE_PORT=8888,
        TEST_DATABASE_USER="built-test-user",
        TEST_DATABASE_PSWD="built-test-pswd",
        TEST_DATABASE_HOST="built-test-host",
        TEST_DATABASE_NAME="built-test-name",
        TEST_DATABASE_PORT=9999,
    ):
        assert build_db_url() == (
            "postgresql://built-base-user:built-base-pswd@built-base-host:8888/built-base-name"
        )


def test_build_db_url__uses_TEST_prefixed_database_settings_if_passed_the_force_test_flag(tweak_settings):
    """
    Provide a test ase for the ``build_db_url()`` function.

    Tests that the build_db_url function computes a database url value from separate
    TEST_DATABASE_ settings when the DEPLOY_ENV is TEST.
    """
    with tweak_settings(
        DEPLOY_ENV="TEST",
        DATABASE_USER="built-base-user",
        DATABASE_PSWD="built-base-pswd",
        DATABASE_HOST="built-base-host",
        DATABASE_NAME="built-base-name",
        DATABASE_PORT=8888,
        TEST_DATABASE_USER="built-test-user",
        TEST_DATABASE_PSWD="built-test-pswd",
        TEST_DATABASE_HOST="built-test-host",
        TEST_DATABASE_NAME="built-test-name",
        TEST_DATABASE_PORT=9999,
    ):
        assert build_db_url() == (
            "postgresql://built-test-user:built-test-pswd@built-test-host:9999/built-test-name"
        )


@pytest.mark.asyncio
async def test_sort_clause__auto_sort_enum_column(database_engine):
    """
    Provide a test case for the ``sort_clause()`` function.

    Test that the sort_clause() function will correctly sort enum columns.
    """

    class DummyStatusEnum(str, enum.Enum):
        CREATED = "CREATED"
        SUBMITTED = "SUBMITTED"
        COMPLETED = "COMPLETED"
        FAILED = "FAILED"
        UNKNOWN = "UNKNOWN"
        REJECTED = "REJECTED"

    dummy_table = Table(
        "dummies",
        metadata,
        Column("id", Integer, primary_key=True),
        Column("status", Enum(DummyStatusEnum, sort_key_function=lambda v: str(v.value)), nullable=False),
    )
    dummy_sortables = [dummy_table.c.id, dummy_table.c.status]

    # Note: do not worry about dropping the table. The pytest fixtures will clean it up for us
    dummy_table.create(database_engine)
    await database.execute_many(
        query=dummy_table.insert(),
        values=[
            dict(id=44, status="COMPLETED"),
            dict(id=45, status="FAILED"),
            dict(id=46, status="REJECTED"),
            dict(id=47, status="COMPLETED"),
            dict(id=48, status="CREATED"),
            dict(id=49, status="SUBMITTED"),
            dict(id=50, status="REJECTED"),
            dict(id=51, status="FAILED"),
        ],
    )
    status_sort_clause = sort_clause("status", dummy_sortables, True)
    query = dummy_table.select().order_by(status_sort_clause, "id")
    results = await database.fetch_all(query)
    assert [(d["id"], d["status"]) for d in results] == [
        (44, "COMPLETED"),
        (47, "COMPLETED"),
        (48, "CREATED"),
        (45, "FAILED"),
        (51, "FAILED"),
        (46, "REJECTED"),
        (50, "REJECTED"),
        (49, "SUBMITTED"),
    ]

    status_sort_clause = sort_clause("status", dummy_sortables, False)
    query = dummy_table.select().order_by(status_sort_clause, "id")
    results = await database.fetch_all(query)
    assert [(d["id"], d["status"]) for d in results] == [
        (49, "SUBMITTED"),
        (46, "REJECTED"),
        (50, "REJECTED"),
        (45, "FAILED"),
        (51, "FAILED"),
        (48, "CREATED"),
        (44, "COMPLETED"),
        (47, "COMPLETED"),
    ]



"""
Test the module version.
"""

from importlib.metadata import PackageNotFoundError
from unittest import mock

from jobbergate_api.version import (
    __version__,
    get_version,
    get_version_from_metadata,
    get_version_from_poetry,
)


def test_get_version_from_metadata():
    """
    Test the function get_version_from_metadata.
    """
    assert get_version_from_metadata() == __version__


def test_get_version_from_poetry():
    """
    Test the function get_version_from_poetry.

    It is necessary to replace alpha and beta by a and b when running
    prereleases on poetry 1.1.*, this can be removed for poetry 1.2.*,
    since the version scheme is already normalized.
    """
    poetry_version = get_version_from_poetry().replace("-alpha.", "a").replace("-beta.", "b")
    assert poetry_version == __version__


@mock.patch("jobbergate_api.version.get_version_from_poetry")
@mock.patch("jobbergate_api.version.get_version_from_metadata")
class TestGetVersion:
    """
    Test class to group tests for the function get_version.
    """

    def test_get_version__success(
        self,
        mock_get_version_from_metadata,
        mock_get_version_from_poetry,
    ):
        """
        Test the function get_version.
        """
        desired_version = "1.2.3"
        mock_get_version_from_metadata.return_value = desired_version

        actual_version = get_version()
        assert actual_version == desired_version

        mock_get_version_from_metadata.assert_called_once()
        mock_get_version_from_poetry.assert_not_called()

    def test_get_version__package_not_found(
        self,
        mock_get_version_from_metadata,
        mock_get_version_from_poetry,
    ):
        """
        Test the function get_version when the package is not installed.
        """
        desired_version = "1.2.3"
        mock_get_version_from_poetry.return_value = desired_version
        mock_get_version_from_metadata.side_effect = PackageNotFoundError

        actual_version = get_version()
        assert actual_version == desired_version

        mock_get_version_from_metadata.assert_called_once()
        mock_get_version_from_poetry.assert_called_once()

    def test_get_version__package_and_toml_not_found(
        self,
        mock_get_version_from_metadata,
        mock_get_version_from_poetry,
    ):
        """
        Test the function get_version when the package and the file pyproject.toml are not found.
        """
        mock_get_version_from_metadata.side_effect = PackageNotFoundError
        mock_get_version_from_poetry.side_effect = FileNotFoundError

        assert get_version() == "unknown"

        mock_get_version_from_metadata.assert_called_once()
        mock_get_version_from_poetry.assert_called_once()



"""
Tests for the /job-submissions/ endpoint.
"""
import pathlib

import pytest
from fastapi import status

from jobbergate_api.apps.applications.models import applications_table
from jobbergate_api.apps.job_scripts.job_script_files import JobScriptFiles
from jobbergate_api.apps.job_scripts.models import job_scripts_table
from jobbergate_api.apps.job_submissions.constants import JobSubmissionStatus
from jobbergate_api.apps.job_submissions.models import job_submissions_table
from jobbergate_api.apps.job_submissions.schemas import JobSubmissionResponse
from jobbergate_api.apps.permissions import Permissions
from jobbergate_api.storage import database


@pytest.mark.asyncio
async def test_create_job_submission__with_client_id_in_token(
    fill_application_data,
    fill_job_script_data,
    fill_job_submission_data,
    client,
    inject_security_header,
    time_frame,
):
    """
    Test POST /job-submissions/ correctly creates a job_submission.

    This test proves that a job_submission is successfully created via a POST request to the /job-submissions/
    endpoint. We show this by asserting that the job_submission is created in the database after the post
    request is made, the correct status code (201) is returned. We also show that the ``client_id``
    is pulled from the token and the created job_submission is connected to that client id.
    """
    inserted_application_id = await database.execute(
        query=applications_table.insert(),
        values=fill_application_data(),
    )
    inserted_job_script_id = await database.execute(
        query=job_scripts_table.insert(),
        values=fill_job_script_data(application_id=inserted_application_id),
    )

    inject_security_header(
        "owner1@org.com",
        Permissions.JOB_SUBMISSIONS_EDIT,
        client_id="dummy-cluster-client",
    )
    create_data = fill_job_submission_data(
        job_script_id=inserted_job_script_id,
        job_submission_name="sub1",
        job_submission_owner_email="owner1@org.com",
    )

    # Removed defaults to make sure these are correctly set by other mechanisms
    create_data.pop("status", None)
    create_data.pop("client_id", None)

    with time_frame() as window:
        response = await client.post("/jobbergate/job-submissions/", json=create_data)

    assert response.status_code == status.HTTP_201_CREATED

    count = await database.fetch_all("SELECT COUNT(*) FROM job_submissions")
    assert count[0][0] == 1

    id_rows = await database.fetch_all("SELECT id FROM job_submissions")
    assert len(id_rows) == 1

    job_submission = JobSubmissionResponse(**response.json())
    assert job_submission.id == id_rows[0][0]
    assert job_submission.job_submission_name == "sub1"
    assert job_submission.job_submission_owner_email == "owner1@org.com"
    assert job_submission.job_submission_description is None
    assert job_submission.job_script_id == inserted_job_script_id
    assert job_submission.execution_directory is None
    assert job_submission.client_id == "dummy-cluster-client"
    assert job_submission.status == JobSubmissionStatus.CREATED
    assert job_submission.created_at in window
    assert job_submission.updated_at in window


@pytest.mark.asyncio
async def test_create_job_submission__with_client_id_in_request_body(
    fill_application_data,
    fill_job_script_data,
    fill_job_submission_data,
    client,
    inject_security_header,
    time_frame,
):
    """
    Test POST /job-submissions/ correctly creates a job_submission.

    This test proves that a job_submission is successfully created via a POST request to the /job-submissions/
    endpoint. We show this by asserting that the job_submission is created in the database after the post
    request is made, the correct status code (201) is returned. We also show that the ``client_id``
    in the request body overrides the client id in the token.
    """
    inserted_application_id = await database.execute(
        query=applications_table.insert(),
        values=fill_application_data(),
    )
    inserted_job_script_id = await database.execute(
        query=job_scripts_table.insert(),
        values=fill_job_script_data(application_id=inserted_application_id),
    )

    inject_security_header(
        "owner1@org.com",
        Permissions.JOB_SUBMISSIONS_EDIT,
        client_id="dummy-cluster-client",
    )
    with time_frame() as window:
        response = await client.post(
            "/jobbergate/job-submissions/",
            json=fill_job_submission_data(
                job_script_id=inserted_job_script_id,
                job_submission_name="sub1",
                job_submission_owner_email="owner1@org.com",
                client_id="silly-cluster-client",
            ),
        )

    assert response.status_code == status.HTTP_201_CREATED

    count = await database.fetch_all("SELECT COUNT(*) FROM job_submissions")
    assert count[0][0] == 1

    id_rows = await database.fetch_all("SELECT id FROM job_submissions")
    assert len(id_rows) == 1

    job_submission = JobSubmissionResponse(**response.json())
    assert job_submission.id == id_rows[0][0]
    assert job_submission.job_submission_name == "sub1"
    assert job_submission.job_submission_owner_email == "owner1@org.com"
    assert job_submission.job_submission_description is None
    assert job_submission.job_script_id == inserted_job_script_id
    assert job_submission.client_id == "silly-cluster-client"
    assert job_submission.status == JobSubmissionStatus.CREATED
    assert job_submission.created_at in window
    assert job_submission.updated_at in window


@pytest.mark.asyncio
async def test_create_job_submission__with_execution_directory(
    fill_application_data,
    fill_job_script_data,
    fill_job_submission_data,
    client,
    inject_security_header,
    time_frame,
):
    """
    Test POST /job-submissions/ correctly creates a job_submission with an execution directory.

    This test proves that a job_submission is successfully created via a POST request to the /job-submissions/
    endpoint with an attached execution directory. We show this by asserting that the job_submission is
    created in the database after the post request is made, the correct status code (201) is returned.
    We also show that the ``execution_directory`` is correctly set.
    """
    inserted_application_id = await database.execute(
        query=applications_table.insert(),
        values=fill_application_data(),
    )
    inserted_job_script_id = await database.execute(
        query=job_scripts_table.insert(),
        values=fill_job_script_data(application_id=inserted_application_id),
    )

    inject_security_header(
        "owner1@org.com",
        Permissions.JOB_SUBMISSIONS_EDIT,
        client_id="dummy-cluster-client",
    )
    create_data = fill_job_submission_data(
        job_script_id=inserted_job_script_id,
        job_submission_name="sub1",
        job_submission_owner_email="owner1@org.com",
        execution_directory="/some/fake/path",
    )

    # Removed defaults to make sure these are correctly set by other mechanisms
    create_data.pop("status", None)
    create_data.pop("client_id", None)

    with time_frame() as window:
        response = await client.post("/jobbergate/job-submissions/", json=create_data)

    assert response.status_code == status.HTTP_201_CREATED

    count = await database.fetch_all("SELECT COUNT(*) FROM job_submissions")
    assert count[0][0] == 1

    id_rows = await database.fetch_all("SELECT id FROM job_submissions")
    assert len(id_rows) == 1

    job_submission = JobSubmissionResponse(**response.json())
    assert job_submission.id == id_rows[0][0]
    assert job_submission.job_submission_name == "sub1"
    assert job_submission.job_submission_owner_email == "owner1@org.com"
    assert job_submission.job_submission_description is None
    assert job_submission.job_script_id == inserted_job_script_id
    assert job_submission.execution_directory == pathlib.Path("/some/fake/path")
    assert job_submission.client_id == "dummy-cluster-client"
    assert job_submission.status == JobSubmissionStatus.CREATED
    assert job_submission.created_at in window
    assert job_submission.updated_at in window


@pytest.mark.asyncio
async def test_create_job_submission_without_job_script(
    client,
    fill_job_submission_data,
    inject_security_header,
):
    """
    Test that is not possible to create a job_submission without a job_script.

    This test proves that is not possible to create a job_submission without an existing job_script.
    We show this by trying to create a job_submission without a job_script created before, then assert that
    the job_submission still does not exists in the database, the correct status code (404) is returned.
    """
    inject_security_header("owner1@org.com", Permissions.JOB_SUBMISSIONS_EDIT)
    response = await client.post(
        "/jobbergate/job-submissions/", json=fill_job_submission_data(job_script_id=9999)
    )
    assert response.status_code == status.HTTP_404_NOT_FOUND

    count = await database.fetch_all("SELECT COUNT(*) FROM job_submissions")
    assert count[0][0] == 0


@pytest.mark.asyncio
async def test_create_job_submission_bad_permission(
    client,
    fill_application_data,
    fill_job_script_data,
    fill_job_submission_data,
    inject_security_header,
):
    """
    Test that is not possible to create a job_submission without the permission.

    This test proves that is not possible to create a job_submission using a user without the permission.
    We show this by trying to create a job_submission with a user without permission, then assert that
    the job_submission still does not exists in the database and the correct status code (403) is returned.
    """
    inserted_application_id = await database.execute(
        query=applications_table.insert(),
        values=fill_application_data(),
    )
    inserted_job_script_id = await database.execute(
        query=job_scripts_table.insert(),
        values=fill_job_script_data(application_id=inserted_application_id),
    )

    inject_security_header("owner1@org.com", "INVALID_PERMISSION")
    response = await client.post(
        "/jobbergate/job-submissions/",
        json=fill_job_submission_data(job_script_id=inserted_job_script_id),
    )
    assert response.status_code == status.HTTP_403_FORBIDDEN

    count = await database.fetch_all("SELECT COUNT(*) FROM job_submissions")
    assert count[0][0] == 0


@pytest.mark.asyncio
async def test_create_job_submission_without_client_id(
    fill_application_data,
    fill_job_script_data,
    fill_job_submission_data,
    client,
    inject_security_header,
):
    """
    Test that it is not possible to create a job_submission without a ``client_id``.

    This test proves that it is not possible to create a job_submission without including a
    ``client_id`` in either the request body or embedded in the access token. If none are supplied,
    we assert that a 400 response is returned.k
    """
    inserted_application_id = await database.execute(
        query=applications_table.insert(),
        values=fill_application_data(),
    )
    inserted_job_script_id = await database.execute(
        query=job_scripts_table.insert(),
        values=fill_job_script_data(application_id=inserted_application_id),
    )

    inject_security_header(
        "owner1@org.com",
        Permissions.JOB_SUBMISSIONS_EDIT,
    )
    create_data = fill_job_submission_data(
        job_script_id=inserted_job_script_id,
        job_submission_name="sub1",
        job_submission_owner_email="owner1@org.com",
    )
    create_data.pop("client_id", None)
    response = await client.post(
        "/jobbergate/job-submissions/",
        json=create_data,
    )

    assert response.status_code == status.HTTP_400_BAD_REQUEST

    count = await database.fetch_all("SELECT COUNT(*) FROM job_submissions")
    assert count[0][0] == 0


@pytest.mark.asyncio
async def test_get_job_submission_by_id(
    client,
    fill_application_data,
    fill_job_script_data,
    fill_job_submission_data,
    inject_security_header,
):
    """
    Test GET /job-submissions/<id>.

    This test proves that GET /job-submissions/<id> returns the correct job-submission, owned by
    the user making the request. We show this by asserting that the job_submission data
    returned in the response is equal to the job_submission data that exists in the database
    for the given job_submission id.
    """
    inserted_application_id = await database.execute(
        query=applications_table.insert(),
        values=fill_application_data(),
    )
    inserted_job_script_id = await database.execute(
        query=job_scripts_table.insert(),
        values=fill_job_script_data(application_id=inserted_application_id),
    )
    inserted_job_submission_id = await database.execute(
        query=job_submissions_table.insert(),
        values=fill_job_submission_data(
            job_script_id=inserted_job_script_id,
            job_submission_name="sub1",
            job_submission_owner_email="owner1@org.com",
        ),
    )

    count = await database.fetch_all("SELECT COUNT(*) FROM job_submissions")
    assert count[0][0] == 1

    inject_security_header("owner1@org.com", Permissions.JOB_SUBMISSIONS_VIEW)
    response = await client.get(f"/jobbergate/job-submissions/{inserted_job_submission_id}")
    assert response.status_code == status.HTTP_200_OK

    data = response.json()
    assert data["id"] == inserted_job_submission_id
    assert data["job_submission_name"] == "sub1"
    assert data["job_submission_owner_email"] == "owner1@org.com"
    assert data["job_script_id"] == inserted_job_script_id


@pytest.mark.asyncio
async def test_get_job_submission_by_id_bad_permission(
    client,
    fill_application_data,
    fill_job_script_data,
    fill_job_submission_data,
    inject_security_header,
):
    """
    Test the correct response code is returned when the user don't have the proper permission.

    This test proves that GET /job-submissions/<id> returns the correct response code when the user don't
    have proper permission. We show this by asserting that the status code returned is 403.
    """
    inserted_application_id = await database.execute(
        query=applications_table.insert(),
        values=fill_application_data(),
    )
    inserted_job_script_id = await database.execute(
        query=job_scripts_table.insert(),
        values=fill_job_script_data(application_id=inserted_application_id),
    )
    inserted_job_submission_id = await database.execute(
        query=job_submissions_table.insert(),
        values=fill_job_submission_data(
            job_script_id=inserted_job_script_id,
            job_submission_owner_email="owner1@org.com",
        ),
    )

    inject_security_header("owner1@org.com", "INVALID_PERMISSION")
    response = await client.get(f"/jobbergate/job-submissions/{inserted_job_submission_id}")
    assert response.status_code == status.HTTP_403_FORBIDDEN


@pytest.mark.asyncio
async def test_get_job_submission_by_id_invalid(client, inject_security_header):
    """
    Test the correct response code is returned when a job_submission does not exist.

    This test proves that GET /job-submissions/<id> returns the correct response code when the
    requested job_submission does not exist. We show this by asserting that the status code
    returned is what we would expect given the job_submission requested doesn't exist (404).
    """
    inject_security_header("owner1@org.com", Permissions.JOB_SUBMISSIONS_VIEW)
    response = await client.get("/jobbergate/job-submissions/9999")
    assert response.status_code == status.HTTP_404_NOT_FOUND


@pytest.mark.asyncio
async def test_get_job_submissions__no_param(
    client,
    fill_application_data,
    fill_job_script_data,
    fill_all_job_submission_data,
    inject_security_header,
):
    """
    Test GET /job-submissions/ returns only job_submissions owned by the user making the request.

    This test proves that GET /job-submissions/ returns the correct job_submissions for the user making
    the request. We show this by asserting that the job_submissions returned in the response are
    only job_submissions owned by the user making the request.
    """
    inserted_application_id = await database.execute(
        query=applications_table.insert(),
        values=fill_application_data(),
    )
    inserted_job_script_id = await database.execute(
        query=job_scripts_table.insert(),
        values=fill_job_script_data(application_id=inserted_application_id),
    )
    await database.execute_many(
        query=job_submissions_table.insert(),
        values=fill_all_job_submission_data(
            dict(
                job_script_id=inserted_job_script_id,
                job_submission_name="sub1",
                job_submission_owner_email="owner1@org.com",
            ),
            dict(
                job_script_id=inserted_job_script_id,
                job_submission_name="sub2",
                job_submission_owner_email="owner999@org.com",
            ),
            dict(
                job_script_id=inserted_job_script_id,
                job_submission_name="sub3",
                job_submission_owner_email="owner1@org.com",
            ),
        ),
    )

    count = await database.fetch_all("SELECT COUNT(*) FROM job_submissions")
    assert count[0][0] == 3

    inject_security_header("owner1@org.com", Permissions.JOB_SUBMISSIONS_VIEW)
    response = await client.get("/jobbergate/job-submissions/")
    assert response.status_code == status.HTTP_200_OK

    data = response.json()
    results = data.get("results")
    assert results
    assert [d["job_submission_name"] for d in results] == ["sub1", "sub3"]

    pagination = data.get("pagination")
    assert pagination == dict(
        total=2,
        start=None,
        limit=None,
    )


@pytest.mark.asyncio
async def test_get_job_submissions__bad_permission(
    client,
    fill_application_data,
    fill_job_script_data,
    fill_job_submission_data,
    inject_security_header,
):
    """
    Test GET /job-submissions/ returns 403 for a user without permission.

    This test proves that GET /job-submissions/ returns the correct status code (403) for a user without
    permission. We show this by asserting that the status code of the response is 403.
    """
    inserted_application_id = await database.execute(
        query=applications_table.insert(),
        values=fill_application_data(),
    )
    inserted_job_script_id = await database.execute(
        query=job_scripts_table.insert(),
        values=fill_job_script_data(application_id=inserted_application_id),
    )
    await database.execute(
        query=job_submissions_table.insert(),
        values=fill_job_submission_data(
            job_script_id=inserted_job_script_id,
            job_submission_owner_email="owner1@org.com",
        ),
    )

    count = await database.fetch_all("SELECT COUNT(*) FROM job_submissions")
    assert count[0][0] == 1

    inject_security_header("owner1@org.com", "INVALID_PERMISSION")
    response = await client.get("/jobbergate/job-submissions/")
    assert response.status_code == status.HTTP_403_FORBIDDEN


@pytest.mark.asyncio
async def test_get_job_submissions__with_all_param(
    client,
    fill_application_data,
    fill_job_script_data,
    fill_all_job_submission_data,
    inject_security_header,
):
    """
    Test that listing job_submissions, when all=True, contains job_submissions owned by other users.

    This test proves that the user making the request can see job_submissions owned by other users.
    We show this by creating three job_submissions, one that are owned by the user making the request, and two
    owned by another user. Assert that the response to GET /job-submissions/?all=True includes all three
    job_submissions.
    """
    inserted_application_id = await database.execute(
        query=applications_table.insert(),
        values=fill_application_data(),
    )
    inserted_job_script_id = await database.execute(
        query=job_scripts_table.insert(),
        values=fill_job_script_data(application_id=inserted_application_id),
    )
    await database.execute_many(
        query=job_submissions_table.insert(),
        values=fill_all_job_submission_data(
            dict(
                job_script_id=inserted_job_script_id,
                job_submission_name="sub1",
                job_submission_owner_email="owner1@org.com",
            ),
            dict(
                job_script_id=inserted_job_script_id,
                job_submission_name="sub2",
                job_submission_owner_email="owner999@org.com",
            ),
            dict(
                job_script_id=inserted_job_script_id,
                job_submission_name="sub3",
                job_submission_owner_email="owner1@org.com",
            ),
        ),
    )

    count = await database.fetch_all("SELECT COUNT(*) FROM job_submissions")
    assert count[0][0] == 3

    inject_security_header("owner1@org.com", Permissions.JOB_SUBMISSIONS_VIEW)
    response = await client.get("/jobbergate/job-submissions/?all=True")
    assert response.status_code == status.HTTP_200_OK

    data = response.json()
    results = data.get("results")
    assert results
    assert [d["job_submission_name"] for d in results] == ["sub1", "sub2", "sub3"]

    pagination = data.get("pagination")
    assert pagination == dict(
        total=3,
        start=None,
        limit=None,
    )


@pytest.mark.asyncio
async def test_get_job_submissions__with_status_param(
    client,
    fill_application_data,
    fill_job_script_data,
    fill_all_job_submission_data,
    inject_security_header,
):
    """
    Test that listing job_submissions, when status is set, contains job_submissions with matching status.

    This test proves that job_submissions are filtered by the status parameter. We do this by setting the
    status param and making sure that only job_submissions with that status are returned.
    """
    inserted_application_id = await database.execute(
        query=applications_table.insert(),
        values=fill_application_data(),
    )
    inserted_job_script_id = await database.execute(
        query=job_scripts_table.insert(),
        values=fill_job_script_data(application_id=inserted_application_id),
    )
    await database.execute_many(
        query=job_submissions_table.insert(),
        values=fill_all_job_submission_data(
            dict(
                job_script_id=inserted_job_script_id,
                job_submission_name="sub1",
                job_submission_owner_email="owner1@org.com",
                status=JobSubmissionStatus.CREATED,
            ),
            dict(
                job_script_id=inserted_job_script_id,
                job_submission_name="sub2",
                job_submission_owner_email="owner1@org.com",
                status=JobSubmissionStatus.CREATED,
            ),
            dict(
                job_script_id=inserted_job_script_id,
                job_submission_name="sub3",
                job_submission_owner_email="owner1@org.com",
                status=JobSubmissionStatus.COMPLETED,
            ),
        ),
    )

    count = await database.fetch_all("SELECT COUNT(*) FROM job_submissions")
    assert count[0][0] == 3

    inject_security_header("owner1@org.com", Permissions.JOB_SUBMISSIONS_VIEW)
    response = await client.get(f"/jobbergate/job-submissions/?submit_status={JobSubmissionStatus.CREATED}")
    assert response.status_code == status.HTTP_200_OK

    data = response.json()
    results = data.get("results")
    assert results
    assert [d["job_submission_name"] for d in results] == ["sub1", "sub2"]

    pagination = data.get("pagination")
    assert pagination == dict(
        total=2,
        start=None,
        limit=None,
    )


@pytest.mark.asyncio
async def test_get_job_submissions__with_search_param(
    client,
    inject_security_header,
    fill_application_data,
    fill_job_script_data,
    fill_all_job_submission_data,
):
    """
    Test that listing job submissions, when search=<search terms>, returns matches.

    This test proves that the user making the request will be shown job submissions that match the search
    string.  We show this by creating job submissions and using various search queries to match against them.

    Assert that the response to GET /job_submissions?search=<search temrms> includes correct matches.
    """
    inserted_application_id = await database.execute(
        query=applications_table.insert(),
        values=fill_application_data(),
    )
    inserted_job_script_id = await database.execute(
        query=job_scripts_table.insert(),
        values=fill_job_script_data(application_id=inserted_application_id),
    )
    await database.execute_many(
        query=job_submissions_table.insert(),
        values=fill_all_job_submission_data(
            dict(
                job_script_id=inserted_job_script_id,
                job_submission_name="test name one",
                job_submission_owner_email="one@org.com",
            ),
            dict(
                job_script_id=inserted_job_script_id,
                job_submission_name="test name two",
                job_submission_owner_email="two@org.com",
            ),
            dict(
                job_script_id=inserted_job_script_id,
                job_submission_name="test name twenty-two",
                job_submission_owner_email="twenty-two@org.com",
                job_submission_description="a long description of this job_script",
            ),
        ),
    )
    count = await database.fetch_all("SELECT COUNT(*) FROM job_submissions")
    assert count[0][0] == 3

    inject_security_header("admin@org.com", Permissions.JOB_SUBMISSIONS_VIEW)

    response = await client.get("/jobbergate/job-submissions?all=true&search=one")
    assert response.status_code == status.HTTP_200_OK
    data = response.json()
    results = data.get("results")
    assert [d["job_submission_name"] for d in results] == ["test name one"]

    response = await client.get("/jobbergate/job-submissions?all=true&search=two")
    assert response.status_code == status.HTTP_200_OK
    data = response.json()
    results = data.get("results")
    assert [d["job_submission_name"] for d in results] == ["test name two", "test name twenty-two"]

    response = await client.get("/jobbergate/job-submissions?all=true&search=long")
    assert response.status_code == status.HTTP_200_OK
    data = response.json()
    results = data.get("results")
    assert [d["job_submission_name"] for d in results] == ["test name twenty-two"]

    response = await client.get("/jobbergate/job-submissions?all=true&search=name+test")
    assert response.status_code == status.HTTP_200_OK
    data = response.json()
    results = data.get("results")
    assert [d["job_submission_name"] for d in results] == [
        "test name one",
        "test name two",
        "test name twenty-two",
    ]


@pytest.mark.asyncio
async def test_get_job_submissions_with_sort_params(
    client,
    fill_application_data,
    fill_job_script_data,
    fill_all_job_submission_data,
    inject_security_header,
):
    """
    Test that listing job_submissions with sort params returns correctly ordered matches.

    This test proves that the user making the request will be shown job_submissions sorted in the correct
    order according to the ``sort_field`` and ``sort_ascending`` parameters.
    We show this by creating job_submissions and using various sort parameters to order them.

    Assert that the response to GET /job_submissions?sort_field=<field>&sort_ascending=<bool> includes
    correctly sorted job_submissions.
    """
    inserted_application_id = await database.execute(
        query=applications_table.insert(),
        values=fill_application_data(),
    )
    inserted_job_script_id = await database.execute(
        query=job_scripts_table.insert(),
        values=fill_job_script_data(application_id=inserted_application_id),
    )
    await database.execute_many(
        query=job_submissions_table.insert(),
        values=fill_all_job_submission_data(
            dict(
                job_script_id=inserted_job_script_id,
                job_submission_name="Z",
                job_submission_owner_email="admin@org.com",
                status=JobSubmissionStatus.REJECTED,
            ),
            dict(
                job_script_id=inserted_job_script_id,
                job_submission_name="Y",
                job_submission_owner_email="admin@org.com",
                status=JobSubmissionStatus.COMPLETED,
            ),
            dict(
                job_script_id=inserted_job_script_id,
                job_submission_name="X",
                job_submission_owner_email="admin@org.com",
                status=JobSubmissionStatus.FAILED,
            ),
        ),
    )
    count = await database.fetch_all("SELECT COUNT(*) FROM job_submissions")
    assert count[0][0] == 3

    inject_security_header("admin@org.com", Permissions.JOB_SUBMISSIONS_VIEW)

    response = await client.get("/jobbergate/job-submissions?sort_field=id")
    assert response.status_code == status.HTTP_200_OK
    data = response.json()
    results = data.get("results")
    assert [d["job_submission_name"] for d in results] == ["Z", "Y", "X"]

    response = await client.get("/jobbergate/job-submissions?sort_field=id&sort_ascending=false")
    assert response.status_code == status.HTTP_200_OK
    data = response.json()
    results = data.get("results")
    assert [d["job_submission_name"] for d in results] == ["X", "Y", "Z"]

    response = await client.get("/jobbergate/job-submissions?sort_field=job_submission_name")
    assert response.status_code == status.HTTP_200_OK
    data = response.json()
    results = data.get("results")
    assert [d["job_submission_name"] for d in results] == ["X", "Y", "Z"]

    response = await client.get("/jobbergate/job-submissions?sort_field=status")
    assert response.status_code == status.HTTP_200_OK
    data = response.json()
    results = data.get("results")
    assert [d["job_submission_name"] for d in results] == ["Y", "X", "Z"]

    response = await client.get("/jobbergate/job-submissions?all=true&sort_field=job_submission_description")
    assert response.status_code == status.HTTP_400_BAD_REQUEST
    assert "Invalid sorting column requested" in response.text


@pytest.mark.asyncio
async def test_list_job_submission_pagination(
    client,
    fill_application_data,
    fill_job_script_data,
    fill_all_job_submission_data,
    inject_security_header,
):
    """
    Test that listing job_submissions works with pagination.

    this test proves that the user making the request can see job_submisions paginated.
    We show this by creating three job_submissions and assert that the response is correctly paginated.
    """
    inserted_application_id = await database.execute(
        query=applications_table.insert(),
        values=fill_application_data(application_owner_email="owner1@org.com"),
    )
    inserted_job_script_id = await database.execute(
        query=job_scripts_table.insert(),
        values=fill_job_script_data(application_id=inserted_application_id),
    )
    await database.execute_many(
        query=job_submissions_table.insert(),
        values=fill_all_job_submission_data(
            *[
                dict(
                    job_submission_name=f"sub{i}",
                    job_script_id=inserted_job_script_id,
                    job_submission_owner_email="owner1@org.com",
                )
                for i in range(1, 6)
            ]
        ),
    )

    count = await database.fetch_all("SELECT COUNT(*) FROM job_submissions")
    assert count[0][0] == 5

    inject_security_header("owner1@org.com", Permissions.JOB_SUBMISSIONS_VIEW)
    response = await client.get("/jobbergate/job-submissions/?start=0&limit=1")
    assert response.status_code == status.HTTP_200_OK

    data = response.json()
    results = data.get("results")
    assert results
    assert [d["job_submission_name"] for d in results] == ["sub1"]

    pagination = data.get("pagination")
    assert pagination == dict(
        total=5,
        start=0,
        limit=1,
    )

    response = await client.get("/jobbergate/job-submissions/?start=1&limit=2")
    assert response.status_code == status.HTTP_200_OK

    data = response.json()
    results = data.get("results")
    assert results
    assert [d["job_submission_name"] for d in results] == ["sub3", "sub4"]

    pagination = data.get("pagination")
    assert pagination == dict(
        total=5,
        start=1,
        limit=2,
    )

    response = await client.get("/jobbergate/job-submissions/?start=2&limit=2")
    assert response.status_code == status.HTTP_200_OK

    data = response.json()
    results = data.get("results")
    assert results
    assert [d["job_submission_name"] for d in results] == ["sub5"]

    pagination = data.get("pagination")
    assert pagination == dict(total=5, start=2, limit=2)


@pytest.mark.asyncio
async def test_get_job_submissions_with_slurm_job_ids_param(
    client,
    fill_application_data,
    fill_job_script_data,
    fill_all_job_submission_data,
    inject_security_header,
):
    """
    Test GET /job-submissions/ returns only job_submissions that have one of the supplied slurm job ids.

    This test proves that GET /job-submissions returns the correct job_submissions with matching slurm_job_id.
    We show this by asserting that the job_submissions returned in the response have one of the supplied
    slurm job ids.
    """
    inserted_application_id = await database.execute(
        query=applications_table.insert(),
        values=fill_application_data(),
    )
    inserted_job_script_id = await database.execute(
        query=job_scripts_table.insert(),
        values=fill_job_script_data(application_id=inserted_application_id),
    )
    await database.execute_many(
        query=job_submissions_table.insert(),
        values=fill_all_job_submission_data(
            dict(
                job_submission_name="sub1",
                job_script_id=inserted_job_script_id,
                job_submission_owner_email="owner1@org.com",
                slurm_job_id=101,
            ),
            dict(
                job_submission_name="sub2",
                job_script_id=inserted_job_script_id,
                job_submission_owner_email="owner1@org.com",
                slurm_job_id=102,
            ),
            dict(
                job_submission_name="sub3",
                job_script_id=inserted_job_script_id,
                job_submission_owner_email="owner1@org.com",
                slurm_job_id=103,
            ),
        ),
    )

    count = await database.fetch_all("SELECT COUNT(*) FROM job_submissions")
    assert count[0][0] == 3

    inject_security_header("owner1@org.com", Permissions.JOB_SUBMISSIONS_VIEW)
    response = await client.get("/jobbergate/job-submissions?slurm_job_ids=101,103")
    assert response.status_code == status.HTTP_200_OK

    data = response.json()
    results = data.get("results")
    assert results
    assert [d["job_submission_name"] for d in results] == ["sub1", "sub3"]

    pagination = data.get("pagination")
    assert pagination == dict(
        total=2,
        start=None,
        limit=None,
    )


@pytest.mark.asyncio
async def test_get_job_submissions_applies_no_slurm_filter_if_slurm_job_ids_is_empty(
    client,
    inject_security_header,
    fill_application_data,
    fill_job_script_data,
    fill_all_job_submission_data,
):
    """
    Test GET /job-submissions/ skips filtering on slurm_job_id if the param is an empty string.

    This test proves that GET /job-submissions doesn't use the slurm_job_id filter if it is an empty string.
    We show this by asserting that passing an empty string as a parameter has no effect.
    """
    inserted_application_id = await database.execute(
        query=applications_table.insert(),
        values=fill_application_data(),
    )
    inserted_job_script_id = await database.execute(
        query=job_scripts_table.insert(),
        values=fill_job_script_data(application_id=inserted_application_id),
    )
    await database.execute_many(
        query=job_submissions_table.insert(),
        values=fill_all_job_submission_data(
            dict(
                job_script_id=inserted_job_script_id,
                job_submission_owner_email="owner1@org.com",
                slurm_job_id=101,
            ),
            dict(
                job_script_id=inserted_job_script_id,
                job_submission_owner_email="owner1@org.com",
                slurm_job_id=102,
            ),
            dict(
                job_script_id=inserted_job_script_id,
                job_submission_owner_email="owner1@org.com",
                slurm_job_id=103,
            ),
        ),
    )

    count = await database.fetch_all("SELECT COUNT(*) FROM job_submissions")
    assert count[0][0] == 3

    inject_security_header("owner1@org.com", Permissions.JOB_SUBMISSIONS_VIEW)

    with_empty_param_response = await client.get("/jobbergate/job-submissions?slurm_job_ids=")
    assert with_empty_param_response.status_code == status.HTTP_200_OK

    without_param_response = await client.get("/jobbergate/job-submissions")
    assert without_param_response.status_code == status.HTTP_200_OK

    assert with_empty_param_response.json() == without_param_response.json()


@pytest.mark.asyncio
async def test_get_job_submissions_with_invalid_slurm_job_ids_param(
    client,
    inject_security_header,
):
    """
    Test GET /job-submissions/ returns a 422 if the slurm_job_id parameter is invalid.

    This test proves that GET /job-submissions requires the slurm_job_ids parameter to be a comma-separated
    list of integer slurm job ids.
    """
    inject_security_header("owner1@org.com", Permissions.JOB_SUBMISSIONS_VIEW)
    response = await client.get("/jobbergate/job-submissions?slurm_job_ids=101-103")
    assert response.status_code == status.HTTP_422_UNPROCESSABLE_ENTITY
    assert "Invalid slurm_job_ids" in response.text

    response = await client.get("/jobbergate/job-submissions?slurm_job_ids=one-oh-one")
    assert response.status_code == status.HTTP_422_UNPROCESSABLE_ENTITY
    assert "Invalid slurm_job_ids" in response.text


@pytest.mark.freeze_time
@pytest.mark.asyncio
async def test_update_job_submission__basic(
    client,
    fill_application_data,
    fill_job_script_data,
    fill_job_submission_data,
    inject_security_header,
    time_frame,
):
    """
    Test update job_submission via PUT.

    This test proves that the job_submission values are correctly updated following a PUT request to the
    /job-submissions/<id> endpoint. We show this by assert the response status code to 201, the response data
    corresponds to the updated data, and the data in the database is also updated.
    """
    inserted_application_id = await database.execute(
        query=applications_table.insert(),
        values=fill_application_data(),
    )
    inserted_job_script_id = await database.execute(
        query=job_scripts_table.insert(),
        values=fill_job_script_data(application_id=inserted_application_id),
    )
    inserted_job_submission_id = await database.execute(
        query=job_submissions_table.insert(),
        values=fill_job_submission_data(
            job_script_id=inserted_job_script_id,
            job_submission_owner_email="owner1@org.com",
        ),
    )

    inject_security_header("owner1@org.com", Permissions.JOB_SUBMISSIONS_EDIT)
    with time_frame() as window:
        response = await client.put(
            f"/jobbergate/job-submissions/{inserted_job_submission_id}",
            json=dict(job_submission_name="new name", job_submission_description="new description"),
        )

    assert response.status_code == status.HTTP_200_OK
    data = response.json()

    assert data["job_submission_name"] == "new name"
    assert data["job_submission_description"] == "new description"
    assert data["id"] == inserted_job_submission_id

    query = job_submissions_table.select(job_submissions_table.c.id == inserted_job_submission_id)
    job_submission_data = await database.fetch_one(query)

    assert job_submission_data is not None
    assert job_submission_data["job_submission_name"] == "new name"
    assert job_submission_data["job_submission_description"] == "new description"
    assert job_submission_data["updated_at"] in window


@pytest.mark.freeze_time
@pytest.mark.asyncio
async def test_update_job_submission__with_execution_dir(
    client,
    fill_application_data,
    fill_job_script_data,
    fill_job_submission_data,
    inject_security_header,
    time_frame,
):
    """
    Test update job_submission via PUT can set execution directory.

    This test proves that the job_submission values are correctly updated following a PUT request to the
    /job-submissions/<id> endpoint. We show this by assert the response status code to 201, the response data
    corresponds to the updated data, and the execution_directory is correctly updated.
    """
    inserted_application_id = await database.execute(
        query=applications_table.insert(),
        values=fill_application_data(),
    )
    inserted_job_script_id = await database.execute(
        query=job_scripts_table.insert(),
        values=fill_job_script_data(application_id=inserted_application_id),
    )
    inserted_job_submission_id = await database.execute(
        query=job_submissions_table.insert(),
        values=fill_job_submission_data(
            job_script_id=inserted_job_script_id,
            job_submission_owner_email="owner1@org.com",
        ),
    )

    inject_security_header("owner1@org.com", Permissions.JOB_SUBMISSIONS_EDIT)
    with time_frame() as window:
        response = await client.put(
            f"/jobbergate/job-submissions/{inserted_job_submission_id}",
            json=dict(execution_directory="/some/fake/path"),
        )

    assert response.status_code == status.HTTP_200_OK
    data = response.json()

    assert data["id"] == inserted_job_submission_id
    assert data["execution_directory"] == "/some/fake/path"

    query = job_submissions_table.select(job_submissions_table.c.id == inserted_job_submission_id)
    job_submission_data = await database.fetch_one(query)

    assert job_submission_data is not None
    assert job_submission_data["execution_directory"] == "/some/fake/path"
    assert job_submission_data["updated_at"] in window


@pytest.mark.asyncio
async def test_update_job_submission_not_found(client, inject_security_header):
    """
    Test that it is not possible to update a job_submission not found.

    This test proves that it is not possible to update a job_submission if it is not found. We show this by
    asserting that the response status code of the request is 404.
    """
    inject_security_header("owner1@org.com", Permissions.JOB_SUBMISSIONS_EDIT)
    response = await client.put("/jobbergate/job-submissions/9999", json=dict(job_submission_name="new name"))

    assert response.status_code == status.HTTP_404_NOT_FOUND


@pytest.mark.asyncio
async def test_update_job_submission_bad_permission(
    client,
    fill_application_data,
    fill_job_script_data,
    fill_job_submission_data,
    inject_security_header,
):
    """
    Test that it is not possible to update a job_submission without the permission.

    This test proves that it is not possible to update a job_submission if the user don't have the permission.
    We show this by asserting that the response status code of the request is 403, and that the data stored in
    the database for the job_submission is not updated.
    """
    inserted_application_id = await database.execute(
        query=applications_table.insert(),
        values=fill_application_data(),
    )
    inserted_job_script_id = await database.execute(
        query=job_scripts_table.insert(),
        values=fill_job_script_data(application_id=inserted_application_id),
    )
    inserted_job_submission_id = await database.execute(
        query=job_submissions_table.insert(),
        values=fill_job_submission_data(
            job_script_id=inserted_job_script_id,
            job_submission_name="old name",
            job_submission_owner_email="owner1@org.com",
        ),
    )

    inject_security_header("owner1@org.com", "INVALID_PERMISSION")
    response = await client.put(
        f"/jobbergate/job-submissions/{inserted_job_submission_id}",
        json=dict(job_submission_name="new name"),
    )

    assert response.status_code == status.HTTP_403_FORBIDDEN

    query = job_submissions_table.select(job_submissions_table.c.id == inserted_job_submission_id)
    job_submission = JobSubmissionResponse.parse_obj(await database.fetch_one(query))

    assert job_submission is not None
    assert job_submission.job_submission_name == "old name"


@pytest.mark.asyncio
async def test_delete_job_submission(
    client,
    fill_application_data,
    fill_job_script_data,
    fill_job_submission_data,
    inject_security_header,
):
    """
    Test delete job_submission via DELETE.

    This test proves that a job_submission is successfully deleted via a DELETE request to the
    /job-submissions/<id> endpoint. We show this by asserting that the job_submission no longer exists in
    the database after the request is made and the correct status code is returned (204).
    """
    inserted_application_id = await database.execute(
        query=applications_table.insert(),
        values=fill_application_data(),
    )
    inserted_job_script_id = await database.execute(
        query=job_scripts_table.insert(),
        values=fill_job_script_data(application_id=inserted_application_id),
    )
    inserted_job_submission_id = await database.execute(
        query=job_submissions_table.insert(),
        values=fill_job_submission_data(
            job_script_id=inserted_job_script_id,
            job_submission_owner_email="owner1@org.com",
        ),
    )

    count = await database.fetch_all("SELECT COUNT(*) FROM job_submissions")
    assert count[0][0] == 1

    inject_security_header("owner1@org.com", Permissions.JOB_SUBMISSIONS_EDIT)
    response = await client.delete(f"/jobbergate/job-submissions/{inserted_job_submission_id}")

    assert response.status_code == status.HTTP_204_NO_CONTENT

    count = await database.fetch_all("SELECT COUNT(*) FROM job_submissions")
    assert count[0][0] == 0


@pytest.mark.asyncio
async def test_delete_job_submission_not_found(client, inject_security_header):
    """
    Test that it is not possible to delete a job_submission that is not found.

    This test proves that it is not possible to delete a job_submission if it does not exists. We show this
    by asserting that a 404 response status code is returned.
    """
    inject_security_header("owner1@org.com", Permissions.JOB_SUBMISSIONS_EDIT)
    response = await client.delete("/jobbergate/job-submissions/9999")

    assert response.status_code == status.HTTP_404_NOT_FOUND


@pytest.mark.asyncio
async def test_delete_job_submission_bad_permission(
    client,
    fill_application_data,
    fill_job_script_data,
    fill_job_submission_data,
    inject_security_header,
):
    """
    Test that it is not possible to delete a job_submission with an user without proper permission.

    This test proves that it is not possible to delete a job_submission if the user don't have the permission.
    We show this by asserting that a 403 response status code is returned and the job_submission still exists
    in the database after the request.
    """
    inserted_application_id = await database.execute(
        query=applications_table.insert(),
        values=fill_application_data(),
    )
    inserted_job_script_id = await database.execute(
        query=job_scripts_table.insert(),
        values=fill_job_script_data(application_id=inserted_application_id),
    )
    inserted_job_submission_id = await database.execute(
        query=job_submissions_table.insert(),
        values=fill_job_submission_data(
            job_script_id=inserted_job_script_id,
            job_submission_owner_email="owner1@org.com",
        ),
    )

    count = await database.fetch_all("SELECT COUNT(*) FROM job_submissions")
    assert count[0][0] == 1

    inject_security_header("owner1@org.com", "INVALID_PERMISSION")
    response = await client.delete(f"/jobbergate/job-submissions/{inserted_job_submission_id}")

    assert response.status_code == status.HTTP_403_FORBIDDEN

    count = await database.fetch_all("SELECT COUNT(*) FROM job_submissions")
    assert count[0][0] == 1


@pytest.mark.asyncio
async def test_job_submissions_agent_pending__success(
    client,
    fill_application_data,
    fill_job_script_data,
    fill_all_job_submission_data,
    inject_security_header,
    mocked_file_manager_factory,
):
    """
    Test GET /job-submissions/agent/pending returns only job_submissions owned by the requesting agent.

    This test proves that GET /job-submissions/agent/pending returns the correct job_submissions for the agent
    making the request. We show this by asserting that the job_submissions returned in the response are
    only job_submissions with a ``client_id`` that matches the ``client_id`` found in the request's
    token payload. We also make sure that the response includes job_script_data_as_string,
    as it is fundamental for the integration with the agent.
    """
    inserted_application_id = await database.execute(
        query=applications_table.insert(),
        values=fill_application_data(),
    )
    inserted_job_script_id = await database.execute(
        query=job_scripts_table.insert(),
        values=fill_job_script_data(application_id=inserted_application_id),
    )

    main_file_path = pathlib.Path("jobbergate.py")
    dummy_job_script_files = JobScriptFiles(
        main_file_path=main_file_path, files={main_file_path: "print(__name__)"}
    )
    dummy_job_script_files.write_to_s3(inserted_job_script_id)

    await database.execute_many(
        query=job_submissions_table.insert(),
        values=fill_all_job_submission_data(
            dict(
                job_script_id=inserted_job_script_id,
                job_submission_name="sub1",
                job_submission_owner_email="email1@dummy.com",
                status=JobSubmissionStatus.CREATED,
                client_id="dummy-client",
            ),
            dict(
                job_script_id=inserted_job_script_id,
                job_submission_name="sub2",
                job_submission_owner_email="email2@dummy.com",
                status=JobSubmissionStatus.COMPLETED,
                client_id="dummy-client",
            ),
            dict(
                job_script_id=inserted_job_script_id,
                job_submission_name="sub3",
                job_submission_owner_email="email3@dummy.com",
                status=JobSubmissionStatus.CREATED,
                client_id="silly-client",
            ),
            dict(
                job_script_id=inserted_job_script_id,
                job_submission_name="sub4",
                job_submission_owner_email="email4@dummy.com",
                status=JobSubmissionStatus.CREATED,
                client_id="dummy-client",
            ),
        ),
    )

    count = await database.fetch_all("SELECT COUNT(*) FROM job_submissions")
    assert count[0][0] == 4

    inject_security_header(
        "who@cares.com",
        Permissions.JOB_SUBMISSIONS_VIEW,
        client_id="dummy-client",
    )
    response = await client.get("/jobbergate/job-submissions/agent/pending")

    assert response.status_code == status.HTTP_200_OK

    data = response.json()
    assert sorted(d["job_submission_name"] for d in data) == ["sub1", "sub4"]
    assert sorted(d["job_submission_owner_email"] for d in data) == ["email1@dummy.com", "email4@dummy.com"]
    assert [JobScriptFiles(**d["job_script_files"]) for d in data] == [dummy_job_script_files] * 2


@pytest.mark.asyncio
async def test_job_submissions_agent_pending__missing_job_script_file(
    client,
    fill_application_data,
    fill_job_script_data,
    fill_all_job_submission_data,
    inject_security_header,
    mocked_file_manager_factory,
):
    """
    Test GET /job-submissions/agent/pending returns a 404.

    This test proves that GET /job-submissions/agent/pending returns a 404 status
    if any of the job-script files is not found. It also makes sure that the missing
    id(s) are included in the response to support debugging.
    """
    inserted_application_id = await database.execute(
        query=applications_table.insert(),
        values=fill_application_data(),
    )
    inserted_job_script_id = await database.execute(
        query=job_scripts_table.insert(),
        values=fill_job_script_data(application_id=inserted_application_id),
    )

    await database.execute_many(
        query=job_submissions_table.insert(),
        values=fill_all_job_submission_data(
            dict(
                job_script_id=inserted_job_script_id,
                job_submission_name="sub1",
                job_submission_owner_email="email1@dummy.com",
                status=JobSubmissionStatus.CREATED,
                client_id="dummy-client",
            ),
            dict(
                job_script_id=inserted_job_script_id,
                job_submission_name="sub2",
                job_submission_owner_email="email2@dummy.com",
                status=JobSubmissionStatus.COMPLETED,
                client_id="dummy-client",
            ),
            dict(
                job_script_id=inserted_job_script_id,
                job_submission_name="sub3",
                job_submission_owner_email="email3@dummy.com",
                status=JobSubmissionStatus.CREATED,
                client_id="silly-client",
            ),
            dict(
                job_script_id=inserted_job_script_id,
                job_submission_name="sub4",
                job_submission_owner_email="email4@dummy.com",
                status=JobSubmissionStatus.CREATED,
                client_id="dummy-client",
            ),
        ),
    )

    count = await database.fetch_all("SELECT COUNT(*) FROM job_submissions")
    assert count[0][0] == 4

    inject_security_header(
        "who@cares.com",
        Permissions.JOB_SUBMISSIONS_VIEW,
        client_id="dummy-client",
    )
    response = await client.get("/jobbergate/job-submissions/agent/pending")

    assert response.status_code == status.HTTP_404_NOT_FOUND
    assert f"JobScript file(s) not found, the missing ids are: {inserted_job_script_id}" in response.text


@pytest.mark.asyncio
async def test_job_submissions_agent_pending__returns_400_if_token_does_not_carry_client_id(
    client,
    inject_security_header,
):
    """
    Test GET /job-submissions/agent/pending returns a 400 if the token payload does not include a client_id.

    This test proves that GET /job-submissions/agent/pending returns a 400 status if the access token used
    to query the route does not include a ``client_id``.
    """
    inject_security_header(
        "who@cares.com",
        Permissions.JOB_SUBMISSIONS_VIEW,
    )
    response = await client.get("/jobbergate/job-submissions/agent/pending")
    assert response.status_code == status.HTTP_400_BAD_REQUEST
    assert "token does not contain a `client_id`" in response.text


@pytest.mark.asyncio
async def test_job_submissions_agent_update__success(
    fill_application_data,
    fill_job_script_data,
    fill_all_job_submission_data,
    client,
    inject_security_header,
):
    """
    Test PUT /job-submissions/{job_submission_id} correctly updates a job_submission status.

    This test proves that a job_submission is successfully updated via a PUT request to the
    /job-submissions/{job_submission_id} endpoint. We show this by asserting that the job_submission is
    updated in the database after the post request is made, the correct status code (200) is returned.
    We also show that the ``status`` column is set to the new status value.
    """
    inserted_application_id = await database.execute(
        query=applications_table.insert(),
        values=fill_application_data(),
    )
    inserted_job_script_id = await database.execute(
        query=job_scripts_table.insert(),
        values=fill_job_script_data(application_id=inserted_application_id),
    )
    await database.execute_many(
        query=job_submissions_table.insert(),
        values=fill_all_job_submission_data(
            dict(
                job_script_id=inserted_job_script_id,
                job_submission_name="sub1",
                status=JobSubmissionStatus.CREATED,
                client_id="dummy-client",
                slurm_job_id=None,
            ),
            dict(
                job_script_id=inserted_job_script_id,
                job_submission_name="sub2",
                status=JobSubmissionStatus.COMPLETED,
                client_id="dummy-client",
                slurm_job_id=None,
            ),
            dict(
                job_script_id=inserted_job_script_id,
                job_submission_name="sub3",
                status=JobSubmissionStatus.CREATED,
                client_id="silly-client",
                slurm_job_id=None,
            ),
            dict(
                job_script_id=inserted_job_script_id,
                job_submission_name="sub4",
                status=JobSubmissionStatus.CREATED,
                client_id="dummy-client",
                slurm_job_id=None,
            ),
        ),
    )

    count = await database.fetch_all("SELECT COUNT(*) FROM job_submissions")
    assert count[0][0] == 4

    target_job_submission = await database.fetch_one(
        query=job_submissions_table.select().where(job_submissions_table.c.job_submission_name == "sub1")
    )
    assert target_job_submission is not None
    job_submission_id = target_job_submission["id"]

    inject_security_header(
        "who@cares.com",
        Permissions.JOB_SUBMISSIONS_EDIT,
        client_id="dummy-client",
    )
    response = await client.put(
        f"/jobbergate/job-submissions/agent/{job_submission_id}",
        json=dict(
            new_status=JobSubmissionStatus.SUBMITTED.value,
            slurm_job_id=111,
        ),
    )
    assert response.status_code == status.HTTP_200_OK

    data = response.json()
    assert data["id"] == job_submission_id
    assert data["status"] == JobSubmissionStatus.SUBMITTED
    assert data["slurm_job_id"] == 111
    assert data["execution_directory"] is None


@pytest.mark.asyncio
async def test_job_submissions_agent_update__returns_400_if_token_does_not_carry_client_id(
    client,
    inject_security_header,
):
    """
    Test PUT /job-submissions/agent/{job_submission_id} returns 400 if client_id not in token payload.

    This test proves that PUT /job-submissions/agent/{job_submission_id} returns a 400 status if the access
    token used to query the route does not include a ``client_id``.
    """
    inject_security_header(
        "who@cares.com",
        Permissions.JOB_SUBMISSIONS_EDIT,
    )
    response = await client.put(
        "/jobbergate/job-submissions/agent/1",
        json=dict(
            new_status=JobSubmissionStatus.SUBMITTED,
            slurm_job_id=111,
        ),
    )
    assert response.status_code == status.HTTP_400_BAD_REQUEST
    assert "token does not contain a `client_id`" in response.text


@pytest.mark.asyncio
async def test_job_submissions_agent_active__success(
    client,
    fill_application_data,
    fill_job_script_data,
    fill_all_job_submission_data,
    inject_security_header,
):
    """
    Test GET /job-submissions/agent/active returns only active job_submissions owned by the requesting agent.

    This test proves that GET /job-submissions/agent/active returns the correct job_submissions for the agent
    making the request. We show this by asserting that the job_submissions returned in the response are
    only job_submissions with a ``client_id`` that matches the ``client_id`` found in the request's
    token payload and have a status of ``SUBMITTED``.
    """
    inserted_application_id = await database.execute(
        query=applications_table.insert(),
        values=fill_application_data(),
    )
    inserted_job_script_id = await database.execute(
        query=job_scripts_table.insert(),
        values=fill_job_script_data(application_id=inserted_application_id),
    )
    await database.execute_many(
        query=job_submissions_table.insert(),
        values=fill_all_job_submission_data(
            dict(
                job_script_id=inserted_job_script_id,
                job_submission_name="sub1",
                status=JobSubmissionStatus.CREATED,
                client_id="dummy-client",
                slurm_job_id=11,
            ),
            dict(
                job_script_id=inserted_job_script_id,
                job_submission_name="sub2",
                status=JobSubmissionStatus.SUBMITTED,
                client_id="dummy-client",
                slurm_job_id=22,
            ),
            dict(
                job_script_id=inserted_job_script_id,
                job_submission_name="sub3",
                status=JobSubmissionStatus.SUBMITTED,
                client_id="silly-client",
                slurm_job_id=33,
            ),
            dict(
                job_script_id=inserted_job_script_id,
                job_submission_name="sub4",
                status=JobSubmissionStatus.SUBMITTED,
                client_id="dummy-client",
                slurm_job_id=44,
            ),
        ),
    )

    count = await database.fetch_all("SELECT COUNT(*) FROM job_submissions")
    assert count[0][0] == 4

    inject_security_header(
        "who@cares.com",
        Permissions.JOB_SUBMISSIONS_VIEW,
        client_id="dummy-client",
    )
    response = await client.get("/jobbergate/job-submissions/agent/active")
    assert response.status_code == status.HTTP_200_OK

    data = response.json()
    assert [d["job_submission_name"] for d in data] == ["sub2", "sub4"]
    assert [d["slurm_job_id"] for d in data] == [22, 44]


@pytest.mark.asyncio
async def test_job_submissions_agent_active__returns_400_if_token_does_not_carry_client_id(
    client,
    inject_security_header,
):
    """
    Test GET /job-submissions/agent/active returns a 400 if the token payload does not include a client_id.

    This test proves that GET /job-submissions/agent/active returns a 400 status if the access token used
    to query the route does not include a ``client_id``.
    """
    inject_security_header(
        "who@cares.com",
        Permissions.JOB_SUBMISSIONS_VIEW,
    )
    response = await client.get("/jobbergate/job-submissions/agent/active")
    assert response.status_code == status.HTTP_400_BAD_REQUEST
    assert "token does not contain a `client_id`" in response.text



"""
Test the pagination.
"""

import json

import pytest
from pydantic import ValidationError

from jobbergate_api.apps.applications.models import applications_table
from jobbergate_api.apps.applications.schemas import ApplicationResponse
from jobbergate_api.pagination import Pagination, Response, package_response
from jobbergate_api.storage import database


def test_init_fails_on_invalid_parameters():
    """
    Tests that the parameters are valid or an exception will be raised.
    """
    with pytest.raises(ValidationError, match="ensure this value is greater than or equal to 0"):
        Pagination(start=-1, limit=1)

    with pytest.raises(ValidationError, match="ensure this value is greater than 0"):
        Pagination(start=1, limit=0)


def test_string_conversion():
    """
    Test the pagination as string.
    """
    pagination = Pagination(start=13, limit=21)

    assert str(pagination) == "start=13 limit=21"


def test_dict():
    """
    Test the dict() method on a pagination instance.
    """
    pagination = Pagination(start=13, limit=21)
    assert pagination.dict() == dict(start=13, limit=21)

    pagination = Pagination(start=None, limit=999)
    assert pagination.dict() == dict()


@pytest.mark.asyncio
@database.transaction(force_rollback=True)
async def test_package_response__without_pagination():
    """
    Test the package_response method without pagination.
    """
    await database.execute_many(
        query=applications_table.insert(),
        values=[
            dict(
                application_owner_email=f"owner{i}@org.com",
                application_name=f"app{i}",
            )
            for i in range(1, 6)
        ],
    )

    query = applications_table.select()
    pagination = Pagination()
    raw_response = await package_response(ApplicationResponse, query, pagination)
    response = Response[ApplicationResponse].parse_obj(json.loads(raw_response.body))

    results = response.results
    assert len(results) == 5
    for (i, result) in enumerate(results):
        assert isinstance(result, ApplicationResponse)
        assert result.application_name == f"app{i + 1}"

    assert response.pagination.total == 5
    assert response.pagination.start is None
    assert response.pagination.limit is None


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "start,limit,total",
    [(0, 1, 1), (6, 2, 13), (2, 3, 10), (7, 2, 13)],
)
@database.transaction(force_rollback=True)
async def test_package_response__with_pagination(start, limit, total):
    """
    Test the package_response method with pagination.

    Parameters test pagination at upper bound and lower bound of total
    """
    await database.execute_many(
        query=applications_table.insert(),
        values=[
            dict(
                id=i,
                application_owner_email=f"owner{i}@org.com",
                application_name=f"app{i}",
            )
            for i in range(1, total + 1)
        ],
    )

    query = applications_table.select()
    pagination = Pagination(start=start, limit=limit)
    raw_response = await package_response(ApplicationResponse, query, pagination)
    response = Response[ApplicationResponse].parse_obj(json.loads(raw_response.body))

    results = response.results
    # Clamps the expected count at upper bound
    expected_count = max(0, min(total - start * limit, limit))
    assert len(results) == expected_count
    for (i, result) in enumerate(results):
        assert isinstance(result, ApplicationResponse)
        assert result.application_name == f"app{i + (start * limit) + 1}"

    assert response.pagination
    assert response.pagination.total == total
    assert response.pagination.start == start
    assert response.pagination.limit == limit
